from typing import Any, List, Mapping, Tuple, Union
import numpy as np
from citylearn.building import Building
from citylearn.data import ZERO_DIVISION_PLACEHOLDER
import logging

LOGGER = logging.getLogger()
class RewardFunction:
    r"""Base and default reward function class.

    The default reward is the electricity consumption from the grid at the current time step returned as a negative value.

    Parameters
    ----------
    env_metadata: Mapping[str, Any]:
        General static information about the environment.
    **kwargs : dict
        Other keyword arguments for custom reward calculation.
    """
    
    def __init__(self, env_metadata: Mapping[str, Any], exponent: float = None, **kwargs):
        penalty_coefficient = kwargs.pop('charging_constraint_penalty_coefficient', None)
        self.env_metadata = env_metadata
        self.exponent = exponent
        self.charging_constraint_penalty_coefficient = penalty_coefficient

    @property
    def env_metadata(self) -> Mapping[str, Any]:
        return self._env_metadata

    @env_metadata.setter
    def env_metadata(self, env_metadata: Mapping[str, Any]):
        self._env_metadata = env_metadata
    
    @property
    def central_agent(self) -> bool:
        """Expect 1 central agent to control all buildings."""

        return self.env_metadata['central_agent']
    
    @property
    def exponent(self) -> float:
        return self.__exponent

    @exponent.setter
    def exponent(self, exponent: float):
        self.__exponent = 1.0 if exponent is None else exponent

    @property
    def charging_constraint_penalty_coefficient(self) -> float:
        return getattr(self, '_charging_constraint_penalty_coefficient', 1.0)

    @charging_constraint_penalty_coefficient.setter
    def charging_constraint_penalty_coefficient(self, coefficient: float):
        if coefficient is None:
            self._charging_constraint_penalty_coefficient = 1.0
        else:
            self._charging_constraint_penalty_coefficient = float(coefficient)

    def reset(self):
        """Use to reset variables at the start of an episode."""

        pass

    def calculate(self, observations: List[Mapping[str, Union[int, float]]]) -> List[float]:
        r"""Calculates reward.

        Parameters
        ----------
        observations: List[Mapping[str, Union[int, float]]]
            List of all building observations at current :py:attr:`citylearn.citylearn.CityLearnEnv.
            time_step` that are got from calling :py:meth:`citylearn.building.Building.observations`.

        Returns
        -------
        reward: List[float]
            Reward for transition to current timestep.
        """

        net_electricity_consumption = [o['net_electricity_consumption'] for o in observations]
        reward_list = [-(max(o, 0)**self.exponent) for o in net_electricity_consumption]

        if self.central_agent:
            reward = [sum(reward_list)]
        else:
            reward = reward_list

        return reward
    
class MultiBuildingRewardFunction(RewardFunction):
    def __init__(self, env, reward_functions: dict[str, RewardFunction]):
        self.env = env
        self.reward_functions = reward_functions
        super().__init__(env)

    def calculate(self, observations: list[dict]) -> list[float]:
        rewards = []
        for obs, (building_name, rf) in zip(observations, self.reward_functions.items()):
            if rf is None:
                raise ValueError(f"No reward function for building '{building_name}'")

            rewards.append(rf.calculate([obs]))
        return rewards

    def reset(self):
        for rf in self.reward_functions.values():
            rf.reset()

    @property
    def env_metadata(self):
        return self._env_metadata

    @env_metadata.setter
    def env_metadata(self, env_metadata: Mapping[str, Any]):
        self._env_metadata = env_metadata
        for rf in self.reward_functions.values():
            rf.env_metadata = env_metadata
    

class MARL(RewardFunction):
    """MARL reward function class.

    Parameters
    ----------
    env_metadata: Mapping[str, Any]:
        General static information about the environment.
    """

    def __init__(self, env_metadata: Mapping[str, Any]):
        super().__init__(env_metadata)

    def calculate(self, observations: List[Mapping[str, Union[int, float]]]) -> List[float]:
        net_electricity_consumption = [o['net_electricity_consumption'] for o in observations]
        district_electricity_consumption = sum(net_electricity_consumption)
        building_electricity_consumption = np.array(net_electricity_consumption, dtype=float)*-1
        reward_list = np.sign(building_electricity_consumption)*0.01*building_electricity_consumption**2*np.nanmax([0, district_electricity_consumption])

        if self.central_agent:
            reward = [reward_list.sum()]
        else:
            reward = reward_list.tolist()
        
        return reward

class IndependentSACReward(RewardFunction):
    """Recommended for use with the `SAC` controllers.
    
    Returned reward assumes that the building-agents act independently of each other, without sharing information through the reward.

    Parameters
    ----------
    env_metadata: Mapping[str, Any]:
        General static information about the environment.
    """
    
    def __init__(self, env_metadata: Mapping[str, Any]):
        super().__init__(env_metadata)

    def calculate(self, observations: List[Mapping[str, Union[int, float]]]) -> List[float]:
        net_electricity_consumption = [o['net_electricity_consumption'] for o in observations]
        reward_list = [min(v*-1**3, 0) for v in net_electricity_consumption]

        if self.central_agent:
            reward = [sum(reward_list)]
        else:
            reward = reward_list

        return reward
    
class SolarPenaltyReward(RewardFunction):
    """The reward is designed to minimize electricity consumption and maximize solar generation to charge energy storage systems.

    The reward is calculated for each building, i and summed to provide the agent with a reward that is representative of all the
    building or buildings (in centralized case)it controls. It encourages net-zero energy use by penalizing grid load satisfaction 
    when there is energy in the energy storage systems as well as penalizing net export when the energy storage systems are not
    fully charged through the penalty term. There is neither penalty nor reward when the energy storage systems are fully charged
    during net export to the grid. Whereas, when the energy storage systems are charged to capacity and there is net import from the 
    grid the penalty is maximized.

    Parameters
    ----------
    env_metadata: Mapping[str, Any]:
        General static information about the environment.
    """

    def __init__(self, env_metadata: Mapping[str, Any]):
        super().__init__(env_metadata)

    def calculate(self, observations: List[Mapping[str, Union[int, float]]]) -> List[float]:
        reward_list = []

        for o, m in zip(observations, self.env_metadata['buildings']):
            e = o['net_electricity_consumption']
            cc = m['cooling_storage']['capacity']
            hc = m['heating_storage']['capacity']
            dc = m['dhw_storage']['capacity']
            ec = m['electrical_storage']['capacity']
            cs = o.get('cooling_storage_soc', 0.0)
            hs = o.get('heating_storage_soc', 0.0)
            ds = o.get('dhw_storage_soc', 0.0)
            es = o.get('electrical_storage_soc', 0.0)
            reward = 0.0
            reward += -(1.0 + np.sign(e)*cs)*abs(e) if cc > ZERO_DIVISION_PLACEHOLDER else 0.0
            reward += -(1.0 + np.sign(e)*hs)*abs(e) if hc > ZERO_DIVISION_PLACEHOLDER else 0.0
            reward += -(1.0 + np.sign(e)*ds)*abs(e) if dc > ZERO_DIVISION_PLACEHOLDER else 0.0
            reward += -(1.0 + np.sign(e)*es)*abs(e) if ec > ZERO_DIVISION_PLACEHOLDER else 0.0
            reward_list.append(reward)

        if self.central_agent:
            reward = [sum(reward_list)]
        else:
            reward = reward_list
        
        return reward
    
class ComfortReward(RewardFunction):
    """Reward for occupant thermal comfort satisfaction.

    The reward is calculated as the negative difference between the setpoint and indoor dry-bulb temperature raised to some exponent
    if outside the comfort band. If within the comfort band, the reward is the negative difference when in cooling mode and temperature
    is below the setpoint or when in heating mode and temperature is above the setpoint. The reward is 0 if within the comfort band
    and above the setpoint in cooling mode or below the setpoint and in heating mode.

    Parameters
    ----------
    env_metadata: Mapping[str, Any]:
        General static information about the environment.
    band: float, default: 2.0
        Setpoint comfort band (+/-). If not provided, the comfort band time series defined in the
        building file, or the default time series value of 2.0 is used.
    lower_exponent: float, default = 2.0
        Penalty exponent for when in cooling mode but temperature is above setpoint upper
        boundary or heating mode but temperature is below setpoint lower boundary.
    higher_exponent: float, default = 2.0
        Penalty exponent for when in cooling mode but temperature is below setpoint lower
        boundary or heating mode but temperature is above setpoint upper boundary.
    """
    
    def __init__(self, env_metadata: Mapping[str, Any], band: float = None, lower_exponent: float = None, higher_exponent: float = None):
        super().__init__(env_metadata)
        self.band = band
        self.lower_exponent = lower_exponent
        self.higher_exponent = higher_exponent

    @property
    def band(self) -> float:
        return self.__band
    
    @property
    def lower_exponent(self) -> float:
        return self.__lower_exponent
    
    @property
    def higher_exponent(self) -> float:
        return self.__higher_exponent
    
    @band.setter
    def band(self, band: float):
        self.__band = band

    @lower_exponent.setter
    def lower_exponent(self, lower_exponent: float):
        self.__lower_exponent = 2.0 if lower_exponent is None else lower_exponent

    @higher_exponent.setter
    def higher_exponent(self, higher_exponent: float):
        self.__higher_exponent = 2.0 if higher_exponent is None else higher_exponent

    def calculate(self, observations: List[Mapping[str, Union[int, float]]]) -> List[float]:
        reward_list = []

        for o in observations:
            heating_demand = o.get('heating_demand', 0.0)
            cooling_demand = o.get('cooling_demand', 0.0)
            heating = heating_demand > cooling_demand
            hvac_mode = o['hvac_mode']
            indoor_dry_bulb_temperature = o['indoor_dry_bulb_temperature']

            if hvac_mode in [1, 2]:
                set_point = o['indoor_dry_bulb_temperature_cooling_set_point'] if hvac_mode == 1 else o['indoor_dry_bulb_temperature_heating_set_point']
                band =  self.band if self.band is not None else o['comfort_band']
                lower_bound_comfortable_indoor_dry_bulb_temperature = set_point - band
                upper_bound_comfortable_indoor_dry_bulb_temperature = set_point + band
                delta = abs(indoor_dry_bulb_temperature - set_point)
                
                if indoor_dry_bulb_temperature < lower_bound_comfortable_indoor_dry_bulb_temperature:
                    exponent = self.lower_exponent if hvac_mode == 2 else self.higher_exponent
                    reward = -(delta**exponent)
                
                elif lower_bound_comfortable_indoor_dry_bulb_temperature <= indoor_dry_bulb_temperature < set_point:
                    reward = 0.0 if heating else -delta

                elif set_point <= indoor_dry_bulb_temperature <= upper_bound_comfortable_indoor_dry_bulb_temperature:
                    reward = -delta if heating else 0.0

                else:
                    exponent = self.higher_exponent if heating else self.lower_exponent
                    reward = -(delta**exponent)

            else:
                cooling_set_point = o['indoor_dry_bulb_temperature_cooling_set_point']
                heating_set_point = o['indoor_dry_bulb_temperature_heating_set_point']
                band =  self.band if self.band is not None else o['comfort_band']
                lower_bound_comfortable_indoor_dry_bulb_temperature = heating_set_point - band
                upper_bound_comfortable_indoor_dry_bulb_temperature = cooling_set_point + band
                cooling_delta = indoor_dry_bulb_temperature - cooling_set_point
                heating_delta = indoor_dry_bulb_temperature - heating_set_point

                if indoor_dry_bulb_temperature < lower_bound_comfortable_indoor_dry_bulb_temperature:
                    exponent = self.higher_exponent if not heating else self.lower_exponent
                    reward = -(abs(heating_delta)**exponent)

                elif lower_bound_comfortable_indoor_dry_bulb_temperature <= indoor_dry_bulb_temperature < heating_set_point:
                    reward = -(abs(heating_delta))

                elif heating_set_point <= indoor_dry_bulb_temperature <= cooling_set_point:
                    reward = 0.0

                elif cooling_set_point < indoor_dry_bulb_temperature < upper_bound_comfortable_indoor_dry_bulb_temperature:
                    reward = -(abs(cooling_delta))

                else:
                    exponent = self.higher_exponent if heating else self.lower_exponent
                    reward = -(abs(cooling_delta)**exponent)

            reward_list.append(reward)

        if self.central_agent:
            reward = [sum(reward_list)]

        else:
            reward = reward_list

        return reward
    
class SolarPenaltyAndComfortReward(RewardFunction):
    """Addition of :py:class:`citylearn.reward_function.SolarPenaltyReward` and :py:class:`citylearn.reward_function.ComfortReward`.

    Parameters
    ----------
    env_metadata: Mapping[str, Any]:
        General static information about the environment.
    band: float, default = 2.0
        Setpoint comfort band (+/-). If not provided, the comfort band time series defined in the
        building file, or the default time series value of 2.0 is used.
    lower_exponent: float, default = 2.0
        Penalty exponent for when in cooling mode but temperature is above setpoint upper
        boundary or heating mode but temperature is below setpoint lower boundary.
    higher_exponent: float, default = 3.0
        Penalty exponent for when in cooling mode but temperature is below setpoint lower
        boundary or heating mode but temperature is above setpoint upper boundary.
    coefficients: Tuple, default = (1.0, 1.0)
        Coefficents for `citylearn.reward_function.SolarPenaltyReward` and :py:class:`citylearn.reward_function.ComfortReward` values respectively.
    """
    
    def __init__(self, env_metadata: Mapping[str, Any], band: float = None, lower_exponent: float = None, higher_exponent: float = None, coefficients: Tuple = None):
        self.__functions: List[RewardFunction] = [
            SolarPenaltyReward(env_metadata),
            ComfortReward(env_metadata, band=band, lower_exponent=lower_exponent, higher_exponent=higher_exponent)
        ]
        super().__init__(env_metadata)
        self.coefficients = coefficients

    @property
    def coefficients(self) -> Tuple:
        return self.__coefficients
    
    @RewardFunction.env_metadata.setter
    def env_metadata(self, env_metadata: Mapping[str, Any]) -> Mapping[str, Any]:
        RewardFunction.env_metadata.fset(self, env_metadata)

        for f in self.__functions:
            f.env_metadata = self.env_metadata
    
    @coefficients.setter
    def coefficients(self, coefficients: Tuple):
        coefficients = [1.0]*len(self.__functions) if coefficients is None else coefficients
        assert len(coefficients) == len(self.__functions), f'{type(self).__name__} needs {len(self.__functions)} coefficients.' 
        self.__coefficients = coefficients

    def calculate(self, observations: List[Mapping[str, Union[int, float]]]) -> List[float]:
        reward = np.array([f.calculate(observations) for f in self.__functions], dtype='float32')
        reward = reward*np.reshape(self.coefficients, (len(self.coefficients), 1))
        reward = reward.sum(axis=0).tolist()

        return reward


class Electric_Vehicles_Reward_Function(MARL):
    """
    Reward function for electric vehicle charging behavior in V2G settings.
    Only affects EV-related behavior; other building logic comes from the superclass.
    """

    def __init__(self, env_metadata: Mapping[str, Any], weights: Mapping[str, float] = None):
        super().__init__(env_metadata)

        # Default tunable weights for EV-related reward components
        self.weights = weights or {
            "no_car_charging": -5.0,
            "battery_limits": -2.0,
            "soc_impossible": -10.0,
            "soc_under": -5.0,
            "close_soc": 10.0,
            "self_ev_consumption": 5.0,
            "extra_self_production": 5.0,
        }
        self._last_base_reward_total = 0.0
        self._last_penalty_total = 0.0
        self._last_base_rewards_per_building: List[float] = []
        self._last_penalties_per_building: List[float] = []

    def calculate(self, observations: List[Mapping[str, Union[int, float, dict]]]) -> List[float]:
        current_reward = super().calculate(observations)
        reward_list = []
        base_rewards = []
        penalty_values = []

        for i, o in enumerate(observations):
            ev_info = o.get("electric_vehicles_chargers_dict", {})
            if not ev_info:
                reward = 0
            else:
                if self.central_agent:
                    reward_value = current_reward[0] if isinstance(current_reward, list) else current_reward
                    reward = self.calculate_ev_penalty(o, reward_value)
                else:
                    reward = self.calculate_ev_penalty(o, current_reward[i])

            base_rewards.append(reward)
            violation = float(o.get("charging_constraint_violation_kwh", 0.0) or 0.0)
            penalty = violation * self.charging_constraint_penalty_coefficient if violation > 0.0 else 0.0
            penalty_values.append(penalty)
            reward -= penalty

            reward_list.append(reward)

        self._last_base_rewards_per_building = base_rewards
        self._last_penalties_per_building = penalty_values
        total_reward = [sum(reward_list)] if self.central_agent else reward_list
        self._last_base_reward_total = sum(base_rewards) if self.central_agent else base_rewards
        self._last_penalty_total = sum(penalty_values) if self.central_agent else penalty_values
        LOGGER.debug(f"Calculated EV reward: {total_reward}")
        return total_reward

    def calculate_ev_penalty(self, o: Mapping[str, Union[int, float, dict]], current_reward: float) -> float:

        penalty_total = 0.0
        ev_chargers: dict = o.get("electric_vehicles_chargers_dict", {})
        net_energy_before = o.get("net_electricity_consumption", 0)

        # Bounding the multiplier to avoid extreme scaling
        penalty_multiplier = 1.0 / (1.0 + abs(current_reward))

        for charger_id, data in ev_chargers.items():
            contributions = {k: 0.0 for k in self.weights.keys()}

            if not data["connected"]:
                if data["last_charged_kwh"] and abs(data["last_charged_kwh"]) > 0.1:
                    contributions["no_car_charging"] += self.weights["no_car_charging"] * penalty_multiplier
                LOGGER.debug(f"Charger {charger_id} | EV not connected | Contributions: {contributions}")
                continue

            # Extract values
            soc_prev = data.get("previous_battery_soc")
            soc_now = data.get("battery_soc")
            capacity = data.get("battery_capacity")
            min_capacity = data.get("min_capacity")
            last_charged_kwh = data.get("last_charged_kwh", 0)
            if last_charged_kwh is None:
                last_charged_kwh = 0.0
            required_soc = data.get("required_soc")
            hours_until_departure = data.get("hours_until_departure", 0)
            max_charging_power = data.get("max_charging_power", 0)
            max_discharging_power = data.get("max_discharging_power", 0)

            if soc_prev is None or soc_now is None or capacity is None:
                raise ValueError("Something went wrong, this values should not be none")
                continue
            # Battery limits
            current_energy = soc_prev * capacity + last_charged_kwh
            if current_energy > capacity or current_energy < min_capacity:
                contributions["battery_limits"] += self.weights["battery_limits"] * penalty_multiplier

            # SoC penalties/rewards
            if required_soc is not None:
                soc_diff = soc_now - required_soc
                soc_diff_kWh = soc_diff * capacity

                max_possible_charge = max_charging_power * hours_until_departure
                max_possible_discharge = max_discharging_power * hours_until_departure

                if soc_diff_kWh > max_possible_charge:
                    contributions["soc_impossible"] += self.weights["soc_impossible"] * penalty_multiplier

                if hours_until_departure == 0:
                    if -0.25 < soc_diff <= -0.10:
                        contributions["soc_under"] += 2 * self.weights["soc_under"] * penalty_multiplier
                    elif soc_diff <= -0.25:
                        contributions["soc_under"] += (self.weights["soc_under"] ** 2) * penalty_multiplier
                    elif -0.10 < soc_diff <= 0.10:
                        contributions["close_soc"] += self.weights["close_soc"] * penalty_multiplier

                if abs(soc_diff_kWh) <= max(max_possible_charge, max_possible_discharge):
                    reward_multiplier = 1.0 / (hours_until_departure + 0.1)
                    contributions["close_soc"] += self.weights["close_soc"] * penalty_multiplier * reward_multiplier

            # Self-production reward
            if last_charged_kwh > 0 and net_energy_before < 0:
                contributions["extra_self_production"] += self.weights["extra_self_production"] * penalty_multiplier
            elif last_charged_kwh < 0 and net_energy_before < 0:
                contributions["extra_self_production"] += -0.5 * self.weights["extra_self_production"] * penalty_multiplier

            # Self-consumption reward
            if last_charged_kwh < 0 and net_energy_before > 0:
                contributions["self_ev_consumption"] += self.weights["self_ev_consumption"] * penalty_multiplier
            elif last_charged_kwh > 0 and net_energy_before > 0:
                contributions["self_ev_consumption"] += -0.5 * self.weights["self_ev_consumption"] * penalty_multiplier

            LOGGER.debug(f"Charger {charger_id} | Contributions: {contributions}")
            penalty_total += sum(contributions.values())

        return penalty_total



import inspect
from typing import List, Dict
import numpy as np
from citylearn.base import Environment, EpisodeTracker
from citylearn.electric_vehicle import ElectricVehicle
from citylearn.data import ChargerSimulation
from citylearn.data import ZERO_DIVISION_PLACEHOLDER
np.seterr(divide='ignore', invalid='ignore')

class Charger(Environment):
    def __init__(
            self, episode_tracker: EpisodeTracker, charger_simulation: ChargerSimulation ,charger_id: str = None, efficiency: float = None, max_charging_power: float = None,
            min_charging_power: float = None, max_discharging_power: float = None,  min_discharging_power: float = None, charge_efficiency_curve: Dict[float, float] = None,
            discharge_efficiency_curve: Dict[float, float] = None, connected_electric_vehicle: ElectricVehicle = None, incoming_electric_vehicle: ElectricVehicle = None,
            phase_connection: str = None, time_step_ratio: int = None,
            **kwargs
    ):
        r"""Initializes the `Electric Vehicle Charger` class with the given attributes.

        Parameters
        ----------
        charger_id: str
            Id through which the charger is uniquely identified in the system
        max_charging_power : float, default 50
            Maximum charging power in kW.
        min_charging_power : float, default 0
            Minimum charging power in kW.
        max_discharging_power : float, default 50
            Maximum discharging power in kW.
        min_discharging_power : float, default 0
            Minimum discharging power in kW.
        charge_efficiency_curve : List, none [[0, 0.83],[0.3, 0.83],[0.7, 0.9],[0.8, 0.9],[1, 0.85]]
            Efficiency curve for charging containing power levels and corresponding efficiency values.
        discharge_efficiency_curve : List, none [[0, 0.83],[0.3, 0.83],[0.7, 0.9],[0.8, 0.9],[1, 0.85]]
            Efficiency curve for discharging containing power levels and corresponding efficiency values.

        Other Parameters
        ----------------
        **kwargs : dict
            Other keyword arguments used to initialize super classes.
        """

        self.efficiency = efficiency
        self.charger_id = charger_id
        self.max_charging_power = max_charging_power
        self.min_charging_power = min_charging_power
        self.max_discharging_power = max_discharging_power
        self.min_discharging_power = min_discharging_power
        self.charge_efficiency_curve = charge_efficiency_curve
        self.discharge_efficiency_curve = discharge_efficiency_curve
        self.connected_electric_vehicle = connected_electric_vehicle
        self.incoming_electric_vehicle = incoming_electric_vehicle
        self.phase_connection = phase_connection
        self.charger_simulation = charger_simulation

        arg_spec = inspect.getfullargspec(super().__init__)
        kwargs = {
            key: value for (key, value) in kwargs.items()
            if (key in arg_spec.args or (arg_spec.varkw is not None))
        }
        self.time_step_ratio = time_step_ratio
        seconds_per_time_step = kwargs.get('seconds_per_time_step', 3600)
        self.algorithm_action_based_time_step_hours_ratio = seconds_per_time_step / 3600
        super().__init__(episode_tracker=episode_tracker,time_step_ratio=time_step_ratio
                        ,**kwargs)

    @property
    def charger_simulation(self) -> ChargerSimulation:

        return self.__charger_simulation
    
    @property
    def charger_id(self) -> str:
        """ID of the charger."""

        return self.__charger_id

    @property
    def max_charging_power(self) -> float:
        """Maximum charging power in kW."""

        return self.__max_charging_power

    @property
    def min_charging_power(self) -> float:
        """Minimum charging power in kW."""

        return self.__min_charging_power

    @property
    def max_discharging_power(self) -> float:
        """Maximum discharging power in kW."""

        return self.__max_discharging_power

    @property
    def min_discharging_power(self) -> float:
        """Minimum discharging power in kW."""

        return self.__min_discharging_power

    @property
    def charge_efficiency_curve(self) -> dict:
        """Efficiency curve for charging containing power levels and corresponding efficiency values."""

        return self.__charge_efficiency_curve

    @property
    def discharge_efficiency_curve(self) -> dict:
        """Efficiency curve for discharging containing power levels and corresponding efficiency values."""

        return self.__discharge_efficiency_curve

    @property
    def connected_electric_vehicle(self) -> ElectricVehicle:
        """Electric_Vehicle currently connected to charger"""

        return self.__connected_ev

    @property
    def incoming_electric_vehicle(self) -> ElectricVehicle:
        """Electric_Vehicle incoming to charger"""

        return self.__incoming_ev

    @property
    def efficiency(self) -> float:
        """Technical efficiency."""

        return self.__efficiency

    @property
    def phase_connection(self) -> str:
        """Electrical phase connection assignment."""

        return self.__phase_connection


    @property
    def past_connected_evs(self) -> List[ElectricVehicle]:
        r"""Each timestep with the list of Past connected Evs or None in the case no electric_vehicle was connected """

        return self.__past_connected_evs

    @property
    def past_charging_action_values_kwh(self) -> List[float]:
        r"""Actions given to charge/discharge in [kWh]. Different from the electricity consumption as in this an action can be given but no electric_vehicle being connect it will not consume such energy"""

        return self.__past_charging_action_values_kwh

    @property
    def electricity_consumption(self) -> List[float]:
        r"""Electricity consumption time series."""

        return self.__electricity_consumption
    
    @property
    def time_step_ratio(self) -> float:
        r"""Electricity consumption time series."""

        return self.__time_step_ratio

    @charger_simulation.setter
    def charger_simulation(self, charger_simulation: ChargerSimulation):
        self.__charger_simulation = charger_simulation

    @charger_id.setter
    def charger_id(self, charger_id: str):
        self.__charger_id = charger_id

    @max_charging_power.setter
    def max_charging_power(self, max_charging_power: float):
            if max_charging_power is None:
                self.__max_charging_power = 50.0
            else:
                self.__max_charging_power = max_charging_power

    @min_charging_power.setter
    def min_charging_power(self, min_charging_power: float):
            if min_charging_power is None:
                self.__min_charging_power = 0.0
            else:
                self.__min_charging_power = min_charging_power

    @max_discharging_power.setter
    def max_discharging_power(self, max_discharging_power: float):
            if max_discharging_power is None:
                self.__max_discharging_power = 50.0
            else:
                self.__max_discharging_power = max_discharging_power

    @min_discharging_power.setter
    def min_discharging_power(self, min_discharging_power: float):
            if min_discharging_power is None:
                self.__min_discharging_power = 0.0
            else:
                self.__min_discharging_power = min_discharging_power

    @charge_efficiency_curve.setter
    def charge_efficiency_curve(self, charge_efficiency_curve: List[List[float]]):
        if charge_efficiency_curve is not None:
            self.__charge_efficiency_curve = np.array(charge_efficiency_curve).T
        else:
            self.__charge_efficiency_curve = None

    @discharge_efficiency_curve.setter
    def discharge_efficiency_curve(self, discharge_efficiency_curve: List[List[float]]):
        if discharge_efficiency_curve is not None:
            self.__discharge_efficiency_curve = np.array(discharge_efficiency_curve).T
        else:
            self.__discharge_efficiency_curve = None

    @connected_electric_vehicle.setter
    def connected_electric_vehicle(self, electric_vehicle: ElectricVehicle):
        self.__connected_ev = electric_vehicle

    @incoming_electric_vehicle.setter
    def incoming_electric_vehicle(self, electric_vehicle: ElectricVehicle):
        self.__incoming_ev = electric_vehicle

    @phase_connection.setter
    def phase_connection(self, phase_connection: str):
        self.__phase_connection = None if phase_connection is None else str(phase_connection)

    @time_step_ratio.setter
    def time_step_ratio(self, time_step_ratio: float):
        self.__time_step_ratio = time_step_ratio    

    @efficiency.setter
    def efficiency(self, efficiency: float):
        if efficiency is None:
            self.__efficiency = 1.0
        else:
            assert 0 < efficiency <= 1, 'efficiency must be > 0.'
            self.__efficiency = efficiency

    def plug_car(self, electric_vehicle: ElectricVehicle):
        """
        Connects a electric_vehicle to the charger.

        Parameters
        ----------
        electric_vehicle : object
            electric_vehicle instance to be connected to the charger.
        """

        if self.connected_electric_vehicle is not None:
            raise ValueError("Charger is already in use.")
        self.connected_electric_vehicle = electric_vehicle
        self.__past_connected_evs[self.time_step] = self.connected_electric_vehicle

    def associate_incoming_car(self, electric_vehicle: ElectricVehicle):
        """
        Associates incoming electric_vehicle to the charger.

        Parameters
        ----------
        electric_vehicle : object
            electric_vehicle instance to be connected to the charger.
        """

        self.incoming_electric_vehicle = electric_vehicle

    def get_efficiency(self, power: float, charging: bool) -> float:
        """
        Returns the efficiency corresponding to a given power level.

        If no efficiency curve is set, returns self.efficiency.
        If a curve is set, interpolates efficiency from the appropriate curve.

        Parameters
        ----------
        power : float
            The charging or discharging power level (normalized between 0 and 1).
        charging : bool
            Whether the power level corresponds to charging (True) or discharging (False).

        Returns
        -------
        float
            The interpolated efficiency at the given power level.
        """
        # Select the correct efficiency curve
        efficiency_curve = self.__charge_efficiency_curve if charging else self.__discharge_efficiency_curve
        # If no curve is set, return default efficiency
        if efficiency_curve is None:
            return self.efficiency  # Default efficiency

        # Ensure efficiency_curve is properly shaped
        assert efficiency_curve.shape[0] == 2, "Efficiency curve must have shape (2, N)."

        power_levels, efficiencies = efficiency_curve  # Unpack rows
        return np.interp(power, power_levels, efficiencies)  # Interpolated efficiency

    def update_connected_electric_vehicle_soc(self, action_value: float):
        """
        Updates the SOC of the connected electric vehicle based on the charging/discharging action.

        Parameters
        ----------
        action_value : float
            The normalized charging or discharging action (range [-1, 1]).
        """
        action_value = float(np.clip(action_value, -1.0, 1.0))

        if action_value == 0.0:
            # Keep EV SOC state consistent at the current timestep even when no power is requested.
            # Without this, SOC at `time_step` can remain the zero-initialized default.
            if self.connected_electric_vehicle is not None:
                self.connected_electric_vehicle.battery.charge(0.0)
            self.__electricity_consumption[self.time_step] = 0
            self.__past_charging_action_values_kwh[self.time_step] = 0
            return

        time_step_hours = self.algorithm_action_based_time_step_hours_ratio
        charging = action_value > 0
        efficiency = self.get_efficiency(abs(action_value), charging)

        if charging:
            requested_power_kw = action_value * self.max_charging_power
            if self.max_charging_power <= 0.0:
                applied_power_kw = 0.0
            else:
                lower_bound_kw = min(self.min_charging_power, self.max_charging_power)
                applied_power_kw = max(min(requested_power_kw, self.max_charging_power), lower_bound_kw)
            commanded_energy_kwh = applied_power_kw * time_step_hours
            battery_energy_kwh = commanded_energy_kwh * efficiency
        else:
            requested_power_kw = abs(action_value) * self.max_discharging_power
            if self.max_discharging_power <= 0.0:
                applied_power_kw = 0.0
            else:
                lower_bound_kw = min(self.min_discharging_power, self.max_discharging_power)
                applied_power_kw = max(min(requested_power_kw, self.max_discharging_power), lower_bound_kw)
            commanded_energy_kwh = -applied_power_kw * time_step_hours
            battery_energy_kwh = commanded_energy_kwh / max(efficiency, ZERO_DIVISION_PLACEHOLDER)

        self.__past_charging_action_values_kwh[self.time_step] = commanded_energy_kwh

        if self.connected_electric_vehicle:
            electric_vehicle = self.connected_electric_vehicle

            # Battery model expects dataset-resolution energy. Convert from control-step kWh when needed.
            ratio = getattr(electric_vehicle.battery, 'time_step_ratio', None)
            battery_command = battery_energy_kwh if ratio in (None, 0) else battery_energy_kwh / ratio
            electric_vehicle.battery.charge(battery_command)

            battery_energy_balance = electric_vehicle.battery.energy_balance[self.time_step]
            self.__electricity_consumption[self.time_step] = (
                battery_energy_balance / max(efficiency, ZERO_DIVISION_PLACEHOLDER)
                if battery_energy_balance >= 0 else battery_energy_balance * efficiency
            )
        else:
            self.__electricity_consumption[self.time_step] = 0


    def next_time_step(self):
        r"""Advance to next `time_step`"""

        self.connected_electric_vehicle = None
        self.incoming_electric_vehicle = None
        super().next_time_step()

    def reset(self):
        """
        Resets the Charger to its initial state by disconnecting all electric_vehicles.
        """
        super().reset()
        self.connected_electric_vehicle = None
        self.incoming_electric_vehicle = None
        self.__electricity_consumption = np.zeros(self.episode_tracker.episode_time_steps, dtype='float32')
        self.__past_charging_action_values_kwh = np.zeros(self.episode_tracker.episode_time_steps, dtype='float32')
        self.__past_connected_evs = [None] * self.episode_tracker.episode_time_steps

    def __str__(self):
        return str(self.as_dict())

    
    def as_dict(self) -> dict:
        """
        Return a dictionary representation of the charger and connected EV data.
        """
        net_consumption = self.electricity_consumption[self.time_step]  # dá sempre 0

        if net_consumption > 0:
            consumption = f"{net_consumption}"
            production = "-1.00"
        else:
            consumption = "-1.00"
            production = f"{abs(net_consumption)}"

        charger_data = {
            "Charger Consumption-kWh": consumption,
            "Charger Production-kWh": production,
            "Incoming EV Name": f"{self.incoming_electric_vehicle.name}" if self.incoming_electric_vehicle else "",
            "Charging Action-kWh": f"{self.past_charging_action_values_kwh[self.time_step]}"
        }

        # Check if EV is connected
        connected_ev = self.connected_electric_vehicle
        incoming_electric_vehicle = self.incoming_electric_vehicle
        if connected_ev:
            ev_data = {
                "EV SOC-%": f"{connected_ev.battery.soc[self.time_step]:.2f}",
                "EV Charger State": self.charger_simulation.electric_vehicle_charger_state[self.time_step],
                "EV Required SOC Departure-%": f"{self.charger_simulation.electric_vehicle_required_soc_departure[self.time_step]}",
                "EV Estimated SOC Arrival-%": f"{self.charger_simulation.electric_vehicle_estimated_soc_arrival[self.time_step]}",
                "EV Arrival Time": f"{self.charger_simulation.electric_vehicle_estimated_arrival_time[self.time_step]}",
                "EV Departure Time": f"{self.charger_simulation.electric_vehicle_departure_time[self.time_step]}",
                "Is EV Connected": True,
                "EV Name": connected_ev.name
            }
        elif incoming_electric_vehicle:
            ev_data = {
                "EV SOC-%": f"{incoming_electric_vehicle.battery.soc[self.time_step]:.2f}",
                "EV Charger State": self.charger_simulation.electric_vehicle_charger_state[self.time_step],
                "EV Required SOC Departure-%": f"{self.charger_simulation.electric_vehicle_required_soc_departure[self.time_step]}",
                "EV Estimated SOC Arrival-%": f"{self.charger_simulation.electric_vehicle_estimated_soc_arrival[self.time_step]}",
                "EV Arrival Time": f"{self.charger_simulation.electric_vehicle_estimated_arrival_time[self.time_step]}",
                "EV Departure Time": f"{self.charger_simulation.electric_vehicle_departure_time[self.time_step]}",
                "Is EV Connected": True,
                "EV Name": incoming_electric_vehicle.name
            }
        else:
            ev_data = {
                "EV SOC": "-1.00",
                "EV Charger State": "-1.00",
                "EV Required SOC Departure-%": "-1.00",
                "EV Estimated SOC Arrival-%": "-1.00",
                "EV Arrival Time": "-1.00",
                "EV Departure Time": "-1.00",
                "Is EV Connected": False,
                "EV Name": ""
            }

        # Merge charger and EV data
        charger_data.update(ev_data)
        return charger_data
    
    def render_simulation_end_data(self) -> dict:
        """
        Return a dictionary containing all simulation data across all time steps.
        The returned dictionary is structured with a general charger name and, for each time step,
        a dictionary with the charger data, EV data (if connected), and electricity consumption.

        Returns
        -------
        result : dict
            A JSON-like dictionary with the charger name and per-time-step data.
        """
        num_steps = self.episode_tracker.episode_time_steps - 1
        
        result = {
            'name': self.charger_id,
            'charger_data': []
        }
        
        for t in range(num_steps):
            time_step_data = {
                'time_step': t,
                'charger_id': self.charger_id,
                'electricity_consumption': self.electricity_consumption[t],
                "Charging Action-kWh": f"{self.past_charging_action_values_kwh[t]}",
                'connected_ev': None,
                'incoming_ev': None
            }
            
            # Include connected EV data if available
            if self.connected_electric_vehicle:
                ev = self.connected_electric_vehicle
                time_step_data['connected_ev'] = {
                    'name': ev.name,
                    'soc': ev.battery.soc[t],
                    'capacity': ev.battery.capacity,
                    'charger_state': ev.electric_vehicle_simulation.electric_vehicle_charger_state[t],
                    'required_soc_departure': ev.electric_vehicle_simulation.electric_vehicle_required_soc_departure[t],
                    'estimated_soc_arrival': ev.electric_vehicle_simulation.electric_vehicle_estimated_soc_arrival[t],
                    'arrival_time': ev.electric_vehicle_simulation.electric_vehicle_estimated_arrival_time[t],
                    'departure_time': ev.electric_vehicle_simulation.electric_vehicle_departure_time[t]
                }
            
            # Include incoming EV data if available
            if self.incoming_electric_vehicle:
                time_step_data['incoming_ev'] = {
                    'name': self.incoming_electric_vehicle.name
                }
            
            result['charger_data'].append(time_step_data)
        
        return result



import logging
from typing import Any, List, Union
import numpy as np
import numpy.typing as npt

try:
    import torch
    import torch.nn as nn
    import torch.optim as optim
except (ModuleNotFoundError, ImportError) as e:
    raise Exception("This functionality requires you to install torch. You can install torch by : pip install torch torchvision, or for more detailed instructions please visit https://pytorch.org.")

from citylearn.agents.rbc import RBC
from citylearn.agents.rlc import RLC
from citylearn.citylearn import CityLearnEnv
from citylearn.preprocessing import Encoder, RemoveFeature
from citylearn.rl import PolicyNetwork, ReplayBuffer, SoftQNetwork

class SAC(RLC):
    def __init__(self, env: CityLearnEnv, **kwargs: Any):
        r"""Custom soft actor-critic algorithm.

        Parameters
        ----------
        env: CityLearnEnv
            CityLearn environment.
        
        Other Parameters
        ----------------
        **kwargs : Any
            Other keyword arguments used to initialize super class.
        """

        super().__init__(env, **kwargs)

        # internally defined
        self.normalized = [False for _ in self.action_space]
        self.soft_q_criterion = nn.SmoothL1Loss()
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.replay_buffer = [ReplayBuffer(int(self.replay_buffer_capacity)) for _ in self.action_space]
        self.soft_q_net1 = [None for _ in self.action_space]
        self.soft_q_net2 = [None for _ in self.action_space]
        self.target_soft_q_net1 = [None for _ in self.action_space]
        self.target_soft_q_net2 = [None for _ in self.action_space]
        self.policy_net = [None for _ in self.action_space]
        self.soft_q_optimizer1 = [None for _ in self.action_space]
        self.soft_q_optimizer2 = [None for _ in self.action_space]
        self.policy_optimizer = [None for _ in self.action_space]
        self.target_entropy = [None for _ in self.action_space]
        self.norm_mean = [None for _ in self.action_space]
        self.norm_std = [None for _ in self.action_space]
        self.r_norm_mean = [None for _ in self.action_space]
        self.r_norm_std = [None for _ in self.action_space]
        self.set_networks()

    def update(self, observations: List[List[float]], actions: List[List[float]], reward: List[float], next_observations: List[List[float]], terminated: bool, truncated: bool):
        r"""Update replay buffer.

        Parameters
        ----------
        observations : List[List[float]]
            Previous time step observations.
        actions : List[List[float]]
            Previous time step actions.
        reward : List[float]
            Current time step reward.
        next_observations : List[List[float]]
            Current time step observations.
        terminated : bool
            Indication that episode has ended.
        truncated : bool
            If episode truncates due to a time limit or a reason that is not defined as part of the task MDP.
        """

        # Run once the regression model has been fitted
        # Normalize all the observations using periodical normalization, one-hot encoding, or -1, 1 scaling. It also removes observations that are not necessary (solar irradiance if there are no solar PV panels).

        for i, (o, a, r, n) in enumerate(zip(observations, actions, reward, next_observations)):
            o = self.get_encoded_observations(i, o)
            n = self.get_encoded_observations(i, n)

            if self.normalized[i]:
                o = self.get_normalized_observations(i, o)
                n = self.get_normalized_observations(i, n)
                r = self.get_normalized_reward(i, r)
            else:
                pass
        
            self.replay_buffer[i].push(o, a, r, n, terminated)

            if self.time_step >= self.standardize_start_time_step and self.batch_size <= len(self.replay_buffer[i]):
                if not self.normalized[i]:
                    # calculate normalized observations and rewards
                    X = np.array([j[0] for j in self.replay_buffer[i].buffer], dtype = float)
                    self.norm_mean[i] = np.nanmean(X, axis=0)
                    self.norm_std[i] = np.nanstd(X, axis=0) + 1e-5
                    R = np.array([j[2] for j in self.replay_buffer[i].buffer], dtype = float)
                    self.r_norm_mean[i] = np.nanmean(R, dtype = float)
                    self.r_norm_std[i] = np.nanstd(R, dtype = float)/self.reward_scaling + 1e-5
                    
                    # update buffer with normalization
                    self.replay_buffer[i].buffer = [(
                        np.hstack(self.get_normalized_observations(i, o).reshape(1,-1)[0]),
                        a,
                        self.get_normalized_reward(i, r),
                        np.hstack(self.get_normalized_observations(i, n).reshape(1,-1)[0]),
                        d
                    ) for o, a, r, n, d in self.replay_buffer[i].buffer]
                    self.normalized[i] = True
                
                else:
                    pass

                for _ in range(self.update_per_time_step):
                    o, a, r, n, d = self.replay_buffer[i].sample(self.batch_size)
                    tensor = torch.cuda.FloatTensor if self.device.type == 'cuda' else torch.FloatTensor
                    o = tensor(o).to(self.device)
                    n = tensor(n).to(self.device)
                    a = tensor(a).to(self.device)
                    r = tensor(r).unsqueeze(1).to(self.device)
                    d = tensor(d).unsqueeze(1).to(self.device)

                    with torch.no_grad():
                        # Update Q-values. First, sample an action from the Gaussian policy/distribution for the current (next) observation and its associated log probability of occurrence.
                        new_next_actions, new_log_pi, _ = self.policy_net[i].sample(n)

                        # The updated Q-value is found by subtracting the logprob of the sampled action (proportional to the entropy) to the Q-values estimated by the target networks.
                        target_q_values = torch.min(
                            self.target_soft_q_net1[i](n, new_next_actions),
                            self.target_soft_q_net2[i](n, new_next_actions),
                        ) - self.alpha*new_log_pi
                        q_target = r + (1 - d)*self.discount*target_q_values

                    # Update Soft Q-Networks
                    q1_pred = self.soft_q_net1[i](o, a)
                    q2_pred = self.soft_q_net2[i](o, a)
                    q1_loss = self.soft_q_criterion(q1_pred, q_target)
                    q2_loss = self.soft_q_criterion(q2_pred, q_target)
                    self.soft_q_optimizer1[i].zero_grad()
                    q1_loss.backward()
                    self.soft_q_optimizer1[i].step()
                    self.soft_q_optimizer2[i].zero_grad()
                    q2_loss.backward()
                    self.soft_q_optimizer2[i].step()

                    # Update Policy
                    new_actions, log_pi, _ = self.policy_net[i].sample(o)
                    q_new_actions = torch.min(
                        self.soft_q_net1[i](o, new_actions),
                        self.soft_q_net2[i](o, new_actions)
                    )
                    policy_loss = (self.alpha*log_pi - q_new_actions).mean()
                    self.policy_optimizer[i].zero_grad()
                    policy_loss.backward()
                    self.policy_optimizer[i].step()

                    # Soft Updates
                    for target_param, param in zip(self.target_soft_q_net1[i].parameters(), self.soft_q_net1[i].parameters()):
                        target_param.data.copy_(target_param.data*(1.0 - self.tau) + param.data*self.tau)

                    for target_param, param in zip(self.target_soft_q_net2[i].parameters(), self.soft_q_net2[i].parameters()):
                        target_param.data.copy_(target_param.data*(1.0 - self.tau) + param.data*self.tau)

            else:
                pass

    def predict(self, observations: List[List[float]], deterministic: bool = None):
        r"""Provide actions for current time step.

        Will return randomly sampled actions from `action_space` if :attr:`end_exploration_time_step` <= :attr:`time_step` 
        else will use policy to sample actions.

        Parameters
        ----------
        observations: List[List[float]]
            Environment observations
        deterministic: bool, default: False
            Wether to return purely exploitatative deterministic actions.

        Returns
        -------
        actions: List[float]
            Action values
        """

        deterministic = False if deterministic is None else deterministic

        if self.time_step > self.end_exploration_time_step or deterministic:
            actions = self.get_post_exploration_prediction(observations, deterministic)
            
        else:
            actions = self.get_exploration_prediction(observations)

        self.actions = actions
        self.next_time_step()
        return actions

    def get_post_exploration_prediction(self, observations: List[List[float]], deterministic: bool) -> List[List[float]]:
        """Action sampling using policy, post-exploration time step"""

        actions = []

        for i, o in enumerate(observations):
            o = self.get_encoded_observations(i, o)
            o = self.get_normalized_observations(i, o)
            o = torch.FloatTensor(o).unsqueeze(0).to(self.device)
            result = self.policy_net[i].sample(o)
            a = result[2] if deterministic else result[0]
            actions.append(a.detach().cpu().numpy()[0])

        return actions
            
    def get_exploration_prediction(self, observations: List[List[float]]) -> List[List[float]]:
        """Return randomly sampled actions from `action_space` multiplied by :attr:`action_scaling_coefficient`."""

        # random actions
        return [list(self.action_scaling_coefficient*s.sample()) for s in self.action_space]

    def get_normalized_reward(self, index: int, reward: float) -> float:
        return (reward - self.r_norm_mean[index])/self.r_norm_std[index]

    def get_normalized_observations(self, index: int, observations: List[float]) -> npt.NDArray[np.float64]:
        try:
            return (np.array(observations, dtype = float) - self.norm_mean[index])/self.norm_std[index]
        except:
            # self.time_step >= self.standardize_start_time_step and self.batch_size <= len(self.replay_buffer[i])
            logging.debug('obs:',observations)
            logging.debug('mean:',self.norm_mean[index])
            logging.debug('std:',self.norm_std[index])
            logging.debug(self.time_step, self.standardize_start_time_step, self.batch_size, len(self.replay_buffer[0]))
            assert False

    def get_encoded_observations(self, index: int, observations: List[float]) -> npt.NDArray[np.float64]:
        return np.array([j for j in np.hstack(self.encoders[index]*np.array(observations, dtype=float)) if j != None], dtype = float)

    def set_networks(self, internal_observation_count: int = None):
        internal_observation_count = 0 if internal_observation_count is None else internal_observation_count

        for i in range(len(self.action_dimension)):
            observation_dimension = self.observation_dimension[i] + internal_observation_count
            # init networks
            self.soft_q_net1[i] = SoftQNetwork(observation_dimension, self.action_dimension[i], self.hidden_dimension).to(self.device)
            self.soft_q_net2[i] = SoftQNetwork(observation_dimension, self.action_dimension[i], self.hidden_dimension).to(self.device)
            self.target_soft_q_net1[i] = SoftQNetwork(observation_dimension, self.action_dimension[i], self.hidden_dimension).to(self.device)
            self.target_soft_q_net2[i] = SoftQNetwork(observation_dimension, self.action_dimension[i], self.hidden_dimension).to(self.device)

            for target_param, param in zip(self.target_soft_q_net1[i].parameters(), self.soft_q_net1[i].parameters()):
                target_param.data.copy_(param.data)

            for target_param, param in zip(self.target_soft_q_net2[i].parameters(), self.soft_q_net2[i].parameters()):
                target_param.data.copy_(param.data)

            # Policy
            self.policy_net[i] = PolicyNetwork(observation_dimension, self.action_dimension[i], self.action_space[i], self.action_scaling_coefficient, self.hidden_dimension).to(self.device)
            self.soft_q_optimizer1[i] = optim.Adam(self.soft_q_net1[i].parameters(), lr=self.lr)
            self.soft_q_optimizer2[i] = optim.Adam(self.soft_q_net2[i].parameters(), lr=self.lr)
            self.policy_optimizer[i] = optim.Adam(self.policy_net[i].parameters(), lr=self.lr)
            self.target_entropy[i] = -np.prod(self.action_space[i].shape).item()

    def set_encoders(self) -> List[List[Encoder]]:
        encoders = super().set_encoders()

        for i, o in enumerate(self.observation_names):
            for j, n in enumerate(o):
                if n == 'net_electricity_consumption':
                    encoders[i][j] = RemoveFeature()
            
                else:
                    pass

        return encoders

class SACRBC(SAC):
    r"""Uses :py:class:`citylearn.agents.rbc.RBC` to select actions during exploration before using :py:class:`citylearn.agents.sac.SAC`.

    Parameters
    ----------
    env: CityLearnEnv
        CityLearn environment.
    rbc: RBC
        :py:class:`citylearn.agents.rbc.RBC` or child class, used to select actions during exploration.
    
    Other Parameters
    ----------------
    **kwargs : Any
        Other keyword arguments used to initialize super class.
    """
    
    def __init__(self, env: CityLearnEnv, rbc: Union[RBC, str] = None, **kwargs: Any):
        super().__init__(env, **kwargs)
        self.__set_rbc(rbc, **kwargs)

    @property
    def rbc(self) -> RBC:
        """:py:class:`citylearn.agents.rbc.RBC` class child class or string path to an RBC 
        class e.g. 'citylearn.agents.rbc.RBC', used to select actions during exploration."""

        return self.__rbc
    
    def __set_rbc(self, rbc: RBC, **kwargs):
        if rbc is None:
            rbc = RBC(self.env, **kwargs)
        
        elif isinstance(rbc, RBC):
            pass

        elif isinstance(rbc, str):
            rbc = self.env.load_agent(rbc, env=self.env, **kwargs)

        else:
            rbc = rbc(self.env, **kwargs)
        
        self.__rbc = rbc

    def get_exploration_prediction(self, observations: List[float]) -> List[float]:
        """Return actions using :class:`RBC`."""

        return self.rbc.predict(observations)


from typing import Any, Mapping, Tuple
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader
from citylearn.end_use_load_profiles.lstm_model.model import LSTM
from citylearn.end_use_load_profiles.lstm_model.preprocessing import preprocess_df

def run(config: Mapping[str, Any], df: pd.DataFrame) -> Tuple[LSTM, Mapping[str, Any], Mapping[str, float]]:
    # preprocess data
    data_dict = preprocess_df(config, df)

    # initialize LSTM model
    lstm = LSTM(
        n_features=data_dict['train']['X'].shape[1],
        n_output=data_dict['train']['y'].shape[1],
        seq_len=config['lb'],
        num_layers=config['num_layer'],
        num_hidden=config['hidden_size'],
        drop_prob=config['dropout'],
        weight_decay=config['weight_decay']
    ).to(config['device'])
    criterion = torch.nn.MSELoss()  # mean-squared error for regression
    optimizer = getattr(torch.optim, config['optimizer_name'])(lstm.parameters(), lr=config['learning_rate'])    
    
    # train model
    lstm = train(
        lstm, 
        data_dict['loaders']['train'], 
        data_dict['loaders']['val'],
        optimizer,
        criterion,
        config,
        data_dict['temp_limits']['max'],
        data_dict['temp_limits']['min'],
    )

    # get model error summary
    total_errors = eval(
        config,
        lstm,
        data_dict['loaders']['test'],
        optimizer,
        data_dict['temp_limits']['max'],
        data_dict['temp_limits']['min'],
    )
    
    return lstm, data_dict['observation_metadata'], total_errors

def eval(config: Mapping[str, Any], model: LSTM, test_loader: DataLoader, optimizer: torch.optim.Optimizer, temperature_normalization_maximum: float, temperature_normalization_minimum: float) -> Mapping[str, float]:
    model.eval()
    h = model.init_hidden(config['batch_size'], config['device'])
    ypred = []
    ylab = []
    criteria = [
        torch.nn.L1Loss(), # MAE
        torch.nn.MSELoss(), # RMSE
    ]
    losses = []
    cumulative_error_profiles = []
    
    for batch in test_loader:
        input_test, target_test = batch
        input_test = input_test.to(config['device'])
        target_test = target_test.to(config['device'])
        h = tuple([each.data for each in h])
        output_test, h = model(input_test.float(), h)
        optimizer.zero_grad()
        losses.append([criterion(output_test, target_test.float()).item() for criterion in criteria])
        output_test = output_test.to('cpu')
        output_test = output_test.detach().numpy()
        output_test = output_test[:, 0]
        output_test = output_test * (temperature_normalization_maximum - temperature_normalization_minimum) + temperature_normalization_minimum
        target_test = target_test.to('cpu')
        target_test = target_test.detach().numpy()
        target_test = target_test[:, 0]
        target_test = target_test * (temperature_normalization_maximum - temperature_normalization_minimum) + temperature_normalization_minimum
        ypred.append(output_test)
        ylab.append(target_test)
        cumulative_error_profiles.append((output_test - target_test).cumsum())
    
    ypred, ylab = np.asarray(ypred), np.asarray(ylab)
    absolute_error = (ypred - ylab).mean(axis=1)
    losses = np.asarray(losses)

    return {
        'absolute error': absolute_error.mean(),
        'mae': losses[:, 0].mean(),
        'rmse': losses[:, 1].mean()
    }

def train(lstm: LSTM, train_loader: DataLoader, val_loader: DataLoader, optimizer: torch.optim.Optimizer, criterion: torch.nn.MSELoss, config: Mapping[str, Any], temperature_normalization_maximum: float, temperature_normalization_minimum: float) -> LSTM:
    lstm.train()
    data = []
    train_losses, val_losses = [], []
    
    for _ in range(config['epochs']):
        loss_train, loss_val, \
        ylab_train, ypred_train, \
        ylab_val, ypred_val = _train(model=lstm, train_loader=train_loader, val_loader=val_loader,
                                        optimizer=optimizer, criterion=criterion, config=config,
                                        temperature_normalization_maximum=temperature_normalization_maximum, temperature_normalization_minimum=temperature_normalization_minimum)
        data.append([np.asarray(ylab_val), np.asarray(ypred_val)])
        train_losses.append(loss_train)
        val_losses.append(loss_val)

    return lstm

# TODO: Explicitly define the data type contained in returned lists
def _train(model: LSTM, train_loader: DataLoader, val_loader: DataLoader, optimizer: torch.optim.Optimizer, criterion: torch.nn.MSELoss, config: Mapping[str, Any], temperature_normalization_maximum: float, temperature_normalization_minimum: float, train_loss_list: list = None, val_loss_list: list = None) -> Tuple[list, list, list, list, list, list]:
    ypred_train = []
    ylab_train = []
    ypred_val = []
    ylab_val = []
    train_loss_list = [] if train_loss_list is None else train_loss_list
    val_loss_list = [] if val_loss_list is None else val_loss_list

    model.train()
    # Hidden state is initialized at each epochs
    h = model.init_hidden(config["batch_size"], config["device"])
    for batch in train_loader:
        input_train, target_train = batch
        input_train = input_train.to(config["device"])
        target_train = target_train.to(config["device"])

        # since the batch is big enough, a stateless mode is used
        # (also considering the possibility to shuffle the training examples,
        # which increase the generalization ability of the network)
        h = model.init_hidden(config["batch_size"], config["device"])
        h = tuple([each.data for each in h])

        # FORWARD PASS
        output_train, h = model(input_train.float(), h)
        loss_train = criterion(output_train, target_train.float())

        # BACKWARD PASS
        optimizer.zero_grad()
        # obtain the loss function
        loss_train.backward()

        # STEP WITH OPTIMIZER
        optimizer.step()

        # EVALUATE METRICS FOR TRAIN
        # if config.device.type == 'cuda':
        output_train = output_train.to('cpu')
        output_train = output_train.detach().numpy()
        # RESCALE OUTPUT
        output_train = output_train[:, 0]
        # output_test = np.reshape(output_test, (-1, 1)).shape
        output_train = output_train * (temperature_normalization_maximum - temperature_normalization_minimum) + temperature_normalization_minimum

        # labels = labels.item()
        # if config.device.type == 'cuda':
        target_train = target_train.to('cpu')
        target_train = target_train.detach().numpy()
        target_train = target_train[:, 0]
        # target_test = np.reshape(target_test, (-1, 1))
        # RESCALE LABELS
        target_train = target_train * (temperature_normalization_maximum - temperature_normalization_minimum) + temperature_normalization_minimum
        ypred_train.append(output_train)
        ylab_train.append(target_train)

    # train_loss_list.append(loss_train.item())
    train_loss_list = loss_train.item()

    flatten = lambda l: [item for sublist in l for item in sublist]
    ypred_train = flatten(ypred_train)
    ylab_train = flatten(ylab_train)

    h = model.init_hidden(config["batch_size"], config["device"])
    for batch in val_loader:
        input_val, target_val = batch
        input_val = input_val.to(config["device"])
        target_val = target_val.to(config["device"])

        h = tuple([each.data for each in h])
        output_val, h = model(input_val.float(), h)
        # target_val = target_val.unsqueeze(1)

        optimizer.zero_grad()

        # obtain loss function
        loss_val = criterion(output_val, target_val.float())

        # EVALUATE METRICS FOR TRAIN
        # if config.device.type == 'cuda':
        output_val = output_val.to('cpu')
        output_val = output_val.detach().numpy()
        # RESCALE OUTPUT
        output_val = output_val[:, 0]
        # output_test = np.reshape(output_test, (-1, 1)).shape
        output_val = output_val * (temperature_normalization_maximum - temperature_normalization_minimum) + temperature_normalization_minimum

        # labels = labels.item()
        # if config.device.type == 'cuda':
        target_val = target_val.to('cpu')
        target_val = target_val.detach().numpy()
        target_val = target_val[:, 0]
        # target_test = np.reshape(target_test, (-1, 1))
        # RESCALE LABELS
        target_val = target_val * (temperature_normalization_maximum - temperature_normalization_minimum) + temperature_normalization_minimum
        ypred_val.append(output_val)
        ylab_val.append(target_val)

    # val_loss_list.append(loss_val.item())
    val_loss_list = loss_val.item()

    # if len(train_loss_list) >= (config.epochs - 1):
    #     train_loss_list = []
    # if len(val_loss_list) >= (config.epochs - 1):
    #     val_loss_list = []

    return train_loss_list, val_loss_list, ylab_train, ypred_train, ylab_val, ypred_val


from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, Mapping, Optional, Tuple, Union

import numpy as np

from citylearn.energy_model import HeatPump
from citylearn.preprocessing import Normalize, PeriodicNormalization

if TYPE_CHECKING:
    from citylearn.building import Building

LOGGER = logging.getLogger()


class BuildingOpsService:
    """Internal observation/action operations for `Building`."""

    def __init__(self, building: "Building"):
        self.building = building

    def observations(
        self,
        include_all: bool = None,
        normalize: bool = None,
        periodic_normalization: bool = None,
        check_limits: bool = None,
    ) -> Mapping[str, float]:
        """Observations at current time step."""

        building = self.building

        normalize = False if normalize is None else normalize
        periodic_normalization = False if periodic_normalization is None else periodic_normalization
        include_all = False if include_all is None else include_all
        check_limits = False if check_limits is None else check_limits

        data = self.get_observations_data(include_all=include_all)

        if include_all:
            valid_observations = list(set(data.keys()) | set(building.active_observations))
        else:
            valid_observations = building.active_observations

        observations = {k: data[k] for k in valid_observations if k in data.keys()}

        observations = self.update_ev_charger_observations(
            observations,
            valid_observations,
            building.electric_vehicle_chargers,
            include_all=include_all,
        )

        observations = self.update_washing_machine_observations(
            observations,
            valid_observations,
            building.washing_machines,
        )

        unknown_observations = set(observations.keys()).difference(set(valid_observations))
        assert len(unknown_observations) == 0, f'Unknown observations: {unknown_observations}'

        non_periodic_low_limit, non_periodic_high_limit = building.non_periodic_normalized_observation_space_limits
        periodic_low_limit, periodic_high_limit = building.periodic_normalized_observation_space_limits
        periodic_observations = building.get_periodic_observation_metadata()

        if check_limits:
            for key in building.active_observations:
                value = observations[key]
                lower = non_periodic_low_limit[key]
                upper = non_periodic_high_limit[key]
                if not lower <= value <= upper:
                    report = {
                        'Building': building.name,
                        'episode': building.episode_tracker.episode,
                        'time_step': f'{building.time_step + 1}/{building.episode_tracker.episode_time_steps}',
                        'observation': key,
                        'value': value,
                        'lower': lower,
                        'upper': upper,
                    }
                    LOGGER.debug(f'Observation outside space limit: {report}')

        if periodic_normalization:
            observations_copy = {k: v for k, v in observations.items()}
            observations = {}
            periodic_normalizer = PeriodicNormalization(x_max=0)

            for key, value in observations_copy.items():
                if key in periodic_observations:
                    periodic_normalizer.x_max = max(periodic_observations[key])
                    sin_x, cos_x = value * periodic_normalizer
                    observations[f'{key}_cos'] = cos_x
                    observations[f'{key}_sin'] = sin_x
                else:
                    observations[key] = value

        if normalize:
            normalizer = Normalize(0.0, 1.0)

            for key, value in observations.items():
                normalizer.x_min = periodic_low_limit[key]
                normalizer.x_max = periodic_high_limit[key]
                observations[key] = value * normalizer

        return observations

    def update_ev_charger_observations(self, observations, valid_observations, ev_chargers, include_all: bool = False):
        """Update observations for each electric vehicle charger."""

        building = self.building

        for charger in ev_chargers:
            charger_id = charger.charger_id
            sim = charger.charger_simulation
            t = building.time_step
            endogenous_t = t if include_all else max(t - 1, 0)

            connected_state_key = f'electric_vehicle_charger_{charger_id}_connected_state'
            incoming_state_key = f'electric_vehicle_charger_{charger_id}_incoming_state'
            departure_key = f'connected_electric_vehicle_at_charger_{charger_id}_departure_time'
            req_soc_key = f'connected_electric_vehicle_at_charger_{charger_id}_required_soc_departure'
            soc_key = f'connected_electric_vehicle_at_charger_{charger_id}_soc'
            capacity_key = f'connected_electric_vehicle_at_charger_{charger_id}_battery_capacity'
            arrival_key = f'incoming_electric_vehicle_at_charger_{charger_id}_estimated_arrival_time'
            soc_arrival_key = f'incoming_electric_vehicle_at_charger_{charger_id}_estimated_soc_arrival'

            state = sim.electric_vehicle_charger_state[t] if t < len(sim.electric_vehicle_charger_state) else np.nan

            if charger.connected_electric_vehicle and state == 1:
                if connected_state_key in valid_observations:
                    observations[connected_state_key] = 1
                if departure_key in valid_observations:
                    observations[departure_key] = int(sim.electric_vehicle_departure_time[t])
                if req_soc_key in valid_observations:
                    observations[req_soc_key] = float(sim.electric_vehicle_required_soc_departure[t])
                if soc_key in valid_observations:
                    observations[soc_key] = charger.connected_electric_vehicle.battery.soc[endogenous_t]
                if capacity_key in valid_observations:
                    observations[capacity_key] = float(charger.connected_electric_vehicle.battery.capacity)
            else:
                if connected_state_key in valid_observations:
                    observations[connected_state_key] = 0
                if departure_key in valid_observations:
                    observations[departure_key] = -1
                if req_soc_key in valid_observations:
                    observations[req_soc_key] = -0.1
                if soc_key in valid_observations:
                    observations[soc_key] = -0.1
                if capacity_key in valid_observations:
                    observations[capacity_key] = -1.0

            if charger.incoming_electric_vehicle and state == 2:
                if incoming_state_key in valid_observations:
                    observations[incoming_state_key] = 1
                if arrival_key in valid_observations:
                    observations[arrival_key] = int(sim.electric_vehicle_estimated_arrival_time[t])
                if soc_arrival_key in valid_observations:
                    observations[soc_arrival_key] = float(sim.electric_vehicle_estimated_soc_arrival[t])
            else:
                if incoming_state_key in valid_observations:
                    observations[incoming_state_key] = 0
                if arrival_key in valid_observations:
                    observations[arrival_key] = -1
                if soc_arrival_key in valid_observations:
                    observations[soc_arrival_key] = -0.1

        return observations

    def update_washing_machine_observations(self, observations, valid_observations, washing_machines):
        """Update observations for each washing machine."""

        for washing_machine in washing_machines:
            washing_machine_name = washing_machine.name
            washing_machine_observations = washing_machine.observations()

            start_key = f'{washing_machine_name}_start_time_step'
            if start_key in valid_observations:
                observations[start_key] = next(
                    (value for key, value in washing_machine_observations.items() if '_start_time_step' in key),
                    -1,
                )

            end_key = f'{washing_machine_name}_end_time_step'
            if end_key in valid_observations:
                observations[end_key] = next(
                    (value for key, value in washing_machine_observations.items() if '_end_time_step' in key),
                    -1,
                )
        return observations

    def get_observations_data(self, include_all: bool = False) -> Mapping[str, Union[float, int]]:
        """Build base observation dictionary without normalization."""

        building = self.building

        electric_vehicle_chargers_dict = {}
        washing_machines_dict = {}
        t = building.time_step
        endogenous_t = t if include_all else max(t - 1, 0)

        for charger in building.electric_vehicle_chargers or []:
            charger_id = charger.charger_id
            connected_car = charger.connected_electric_vehicle

            if connected_car is not None:
                last_charged_kwh = 0.0
                if 0 <= endogenous_t < len(charger.past_charging_action_values_kwh):
                    last_charged_kwh = float(charger.past_charging_action_values_kwh[endogenous_t])

                battery_soc = connected_car.battery.soc[endogenous_t]
                previous_battery_soc = connected_car.battery.initial_soc if endogenous_t == 0 else connected_car.battery.soc[endogenous_t - 1]

                required_soc = charger.charger_simulation.electric_vehicle_required_soc_departure[t]
                hours_until_departure = charger.charger_simulation.electric_vehicle_departure_time[t]

                battery_capacity = connected_car.battery.capacity
                min_capacity = (1 - connected_car.battery.depth_of_discharge) * battery_capacity

                electric_vehicle_chargers_dict[charger_id] = {
                    'connected': True,
                    'last_charged_kwh': last_charged_kwh,
                    'previous_battery_soc': previous_battery_soc,
                    'battery_soc': battery_soc,
                    'battery_capacity': battery_capacity,
                    'min_capacity': min_capacity,
                    'required_soc': required_soc,
                    'hours_until_departure': hours_until_departure,
                    'max_charging_power': charger.max_charging_power,
                    'max_discharging_power': charger.max_discharging_power,
                }

            else:
                electric_vehicle_chargers_dict[charger_id] = {
                    'connected': False,
                    'last_charged_kwh': 0.0,
                    'previous_battery_soc': None,
                    'battery_soc': None,
                    'battery_capacity': None,
                    'min_capacity': None,
                    'required_soc': None,
                    'hours_until_departure': None,
                    'max_charging_power': charger.max_charging_power,
                    'max_discharging_power': charger.max_discharging_power,
                }

        for washing_machine in building.washing_machines or []:
            washing_machine_name = washing_machine.name

            def _safe(arr, idx, default):
                try:
                    return arr[idx]
                except Exception:
                    return default

            start_time_step = _safe(washing_machine.washing_machine_simulation.wm_start_time_step, t, -1)
            end_time_step = _safe(washing_machine.washing_machine_simulation.wm_end_time_step, t, -1)
            load_profile = _safe(washing_machine.washing_machine_simulation.load_profile, t, 0.0)

            washing_machines_dict[washing_machine_name] = {
                'wm_start_time_step': start_time_step,
                'wm_end_time_step': end_time_step,
                'load_profile': load_profile,
            }

        observations = {}
        for key, series in building._energy_simulation_observation_sources:
            if t < len(series):
                observations[key] = series[t]

        for key, series in building._weather_observation_sources:
            if t < len(series):
                observations[key] = series[t]

        for key, series in building._pricing_observation_sources:
            if t < len(series):
                observations[key] = series[t]

        for key, series in building._carbon_observation_sources:
            if t < len(series):
                observations[key] = series[t]

        observations.update({
            'solar_generation': abs(building.solar_generation[t]),
            **{
                'cooling_storage_soc': building.cooling_storage.soc[endogenous_t],
                'heating_storage_soc': building.heating_storage.soc[endogenous_t],
                'dhw_storage_soc': building.dhw_storage.soc[endogenous_t],
                'electrical_storage_soc': building.electrical_storage.soc[endogenous_t],
            },
            'cooling_demand': building.energy_from_cooling_device[endogenous_t] + abs(min(building.cooling_storage.energy_balance[endogenous_t], 0.0)),
            'heating_demand': building.energy_from_heating_device[endogenous_t] + abs(min(building.heating_storage.energy_balance[endogenous_t], 0.0)),
            'dhw_demand': building.energy_from_dhw_device[endogenous_t] + abs(min(building.dhw_storage.energy_balance[endogenous_t], 0.0)),
            'net_electricity_consumption': building.net_electricity_consumption[endogenous_t],
            'cooling_electricity_consumption': building.cooling_electricity_consumption[endogenous_t],
            'heating_electricity_consumption': building.heating_electricity_consumption[endogenous_t],
            'dhw_electricity_consumption': building.dhw_electricity_consumption[endogenous_t],
            'cooling_storage_electricity_consumption': building.cooling_storage_electricity_consumption[endogenous_t],
            'heating_storage_electricity_consumption': building.heating_storage_electricity_consumption[endogenous_t],
            'dhw_storage_electricity_consumption': building.dhw_storage_electricity_consumption[endogenous_t],
            'electrical_storage_electricity_consumption': building.electrical_storage_electricity_consumption[endogenous_t],
            'washing_machine_electricity_consumption': building.washing_machines_electricity_consumption[endogenous_t],
            'cooling_device_efficiency': building.cooling_device.get_cop(building.weather.outdoor_dry_bulb_temperature[t], heating=False),
            'heating_device_efficiency': building.heating_device.get_cop(building.weather.outdoor_dry_bulb_temperature[t], heating=True)
            if isinstance(building.heating_device, HeatPump) else building.heating_device.efficiency,
            'dhw_device_efficiency': building.dhw_device.get_cop(building.weather.outdoor_dry_bulb_temperature[t], heating=True)
            if isinstance(building.dhw_device, HeatPump) else building.dhw_device.efficiency,
            'indoor_dry_bulb_temperature_cooling_set_point': building.energy_simulation.indoor_dry_bulb_temperature_cooling_set_point[t],
            'indoor_dry_bulb_temperature_heating_set_point': building.energy_simulation.indoor_dry_bulb_temperature_heating_set_point[t],
            'indoor_dry_bulb_temperature_cooling_delta': building.energy_simulation.indoor_dry_bulb_temperature[t] - building.energy_simulation.indoor_dry_bulb_temperature_cooling_set_point[t],
            'indoor_dry_bulb_temperature_heating_delta': building.energy_simulation.indoor_dry_bulb_temperature[t] - building.energy_simulation.indoor_dry_bulb_temperature_heating_set_point[t],
            'comfort_band': building.energy_simulation.comfort_band[t],
            'occupant_count': building.energy_simulation.occupant_count[t],
            'power_outage': building.power_outage_signal[t],
            'electric_vehicles_chargers_dict': electric_vehicle_chargers_dict,
            'washing_machines_dict': washing_machines_dict,
        })

        if (
            getattr(building, '_charging_constraints_enabled', False)
            and getattr(building, '_expose_charging_constraints', False)
            and isinstance(building._charging_constraints_state, dict)
        ):
            state = building._charging_constraints_state
            headroom = state.get('building_headroom_kw')
            if headroom is not None:
                observations['charging_building_headroom_kw'] = headroom
            export_headroom = state.get('building_export_headroom_kw')
            if export_headroom is not None:
                observations['charging_building_export_headroom_kw'] = export_headroom
            for phase_name, value in (state.get('phase_headroom_kw') or {}).items():
                if value is not None:
                    observations[f'charging_phase_{phase_name}_headroom_kw'] = value
            for phase_name, value in (state.get('phase_export_headroom_kw') or {}).items():
                if value is not None:
                    observations[f'charging_phase_{phase_name}_export_headroom_kw'] = value

        if getattr(building, '_charging_constraints_enabled', False):
            if getattr(building, '_expose_charging_violation', False):
                observations['charging_constraint_violation_kwh'] = building._charging_constraint_last_penalty_kwh
            if getattr(building, '_phase_encoding_observations', None):
                observations.update(building._phase_encoding_observations)

        return observations

    def apply_actions(
        self,
        cooling_or_heating_device_action: float = None,
        cooling_device_action: float = None,
        heating_device_action: float = None,
        cooling_storage_action: float = None,
        heating_storage_action: float = None,
        dhw_storage_action: float = None,
        electrical_storage_action: float = None,
        washing_machine_actions: dict = None,
        electric_vehicle_storage_actions: dict = None,
    ):
        """Update demand and charge/discharge storage devices."""

        building = self.building

        if electric_vehicle_storage_actions is not None:
            electric_vehicle_storage_actions = dict(electric_vehicle_storage_actions)

        if 'cooling_or_heating_device' in building.active_actions:
            assert 'cooling_device' not in building.active_actions and 'heating_device' not in building.active_actions, \
                'cooling_device and heating_device actions must be set to False when cooling_or_heating_device is True.' \
                ' They will be implicitly set based on the polarity of cooling_or_heating_device.'
            cooling_device_action = abs(min(cooling_or_heating_device_action, 0.0))
            heating_device_action = abs(max(cooling_or_heating_device_action, 0.0))

        else:
            assert not ('cooling_device' in building.active_actions and 'heating_device' in building.active_actions), \
                'cooling_device and heating_device actions cannot both be set to True to avoid both actions having' \
                ' values > 0.0 in the same time step. Use cooling_or_heating_device action instead to control' \
                ' both cooling_device and heating_device in a building.'
            cooling_device_action = np.nan if 'cooling_device' not in building.active_actions else cooling_device_action
            heating_device_action = np.nan if 'heating_device' not in building.active_actions else heating_device_action

        cooling_storage_action = 0.0 if 'cooling_storage' not in building.active_actions else cooling_storage_action
        heating_storage_action = 0.0 if 'heating_storage' not in building.active_actions else heating_storage_action
        dhw_storage_action = 0.0 if 'dhw_storage' not in building.active_actions else dhw_storage_action
        electrical_storage_action = 0.0 if 'electrical_storage' not in building.active_actions else electrical_storage_action

        electric_vehicle_storage_actions, electrical_storage_action = self.apply_charging_constraints_to_actions(
            electric_vehicle_storage_actions,
            electrical_storage_action,
        )

        actions = {
            'cooling_demand': (building.update_cooling_demand, (cooling_device_action,)),
            'heating_demand': (building.update_heating_demand, (heating_device_action,)),
            'cooling_device': (building.update_energy_from_cooling_device, ()),
            'cooling_storage': (building.update_cooling_storage, (cooling_storage_action,)),
            'heating_device': (building.update_energy_from_heating_device, ()),
            'heating_storage': (building.update_heating_storage, (heating_storage_action,)),
            'dhw_device': (building.update_energy_from_dhw_device, ()),
            'dhw_storage': (building.update_dhw_storage, (dhw_storage_action,)),
            'non_shiftable_load': (building.update_non_shiftable_load, ()),
            'electrical_storage': (building.update_electrical_storage, (electrical_storage_action,)),
        }

        priority_list = list(actions.keys())

        if electric_vehicle_storage_actions is not None:
            electric_vehicle_priority_list = []
            for charger_id, action in electric_vehicle_storage_actions.items():
                action_key = f'electric_vehicle_storage_{charger_id}'
                if action_key not in building.active_actions:
                    raise ValueError('This action should not be applied. Verify')
                for charger in building.electric_vehicle_chargers:
                    if charger.charger_id == charger_id:
                        actions[action_key] = (charger.update_connected_electric_vehicle_soc, (action,))
                        electric_vehicle_priority_list.append(action_key)
            priority_list = priority_list + electric_vehicle_priority_list

        if washing_machine_actions is not None:
            washing_machine_priority_list = []
            for washing_machine_name, action in washing_machine_actions.items():
                action_key = f'{washing_machine_name}'
                if action_key not in building.active_actions:
                    raise ValueError('This action should not be applied. Verify')
                for washing_machine in building.washing_machines:
                    if washing_machine.name == washing_machine_name:
                        actions[action_key] = (washing_machine.start_cycle, (action,))
                        washing_machine_priority_list.append(action_key)
            priority_list = priority_list + washing_machine_priority_list

        if electrical_storage_action < 0.0:
            key = 'electrical_storage'
            priority_list.remove(key)
            priority_list = [key] + priority_list

        for key in ['cooling', 'heating', 'dhw']:
            storage = f'{key}_storage'
            device = f'{key}_device'

            if actions[storage][1][0] < 0.0:
                storage_ix = priority_list.index(storage)
                device_ix = priority_list.index(device)
                priority_list[storage_ix] = device
                priority_list[device_ix] = storage

        for key in priority_list:
            func, args = actions[key]

            try:
                func(*args)
            except NotImplementedError:
                pass

    def _safe_scalar(self, value, default: float = 0.0) -> float:
        try:
            scalar = float(value)
        except (TypeError, ValueError):
            return float(default)

        if not np.isfinite(scalar):
            return float(default)

        return scalar

    def _safe_index(self, values, idx: int, default: float = 0.0) -> float:
        try:
            return self._safe_scalar(values[idx], default)
        except Exception:
            return float(default)

    def _current_phase_names(self):
        building = self.building
        if getattr(building, '_electrical_service_mode', 'single_phase') == 'three_phase':
            return ['L1', 'L2', 'L3']
        return ['L1']

    def _split_unassigned_power(self, power_kw: float) -> Dict[str, float]:
        building = self.building
        phase_names = self._current_phase_names()

        if len(phase_names) == 1:
            return {'L1': float(power_kw)}

        split_mode = str(getattr(building, '_electrical_service_default_split', 'balanced')).strip().lower()
        if split_mode in {'l1', 'l2', 'l3'}:
            return {phase: float(power_kw if phase.lower() == split_mode else 0.0) for phase in phase_names}

        share = float(power_kw) / len(phase_names)
        return {phase: share for phase in phase_names}

    def _split_power_by_connection(self, power_kw: float, phase_connection: Optional[str]) -> Dict[str, float]:
        phase_names = self._current_phase_names()

        if len(phase_names) == 1:
            return {'L1': float(power_kw)}

        if phase_connection in {'L1', 'L2', 'L3'}:
            return {phase: float(power_kw if phase == phase_connection else 0.0) for phase in phase_names}

        if phase_connection == 'all_phases':
            share = float(power_kw) / len(phase_names)
            return {phase: share for phase in phase_names}

        return self._split_unassigned_power(power_kw)

    def _estimate_non_controllable_base_power(self) -> Tuple[float, Dict[str, float]]:
        building = self.building
        t = building.time_step
        phase_names = self._current_phase_names()

        if building.power_outage:
            return 0.0, {phase: 0.0 for phase in phase_names}

        temperature = self._safe_index(building.weather.outdoor_dry_bulb_temperature, t, 0.0)

        cooling_demand = self._safe_index(building.energy_from_cooling_device, t, 0.0) + self._safe_index(
            building.cooling_storage.energy_balance, t, 0.0
        )
        cooling_kw = self._safe_scalar(building.cooling_device.get_input_power(cooling_demand, temperature, heating=False), 0.0)

        heating_demand = self._safe_index(building.energy_from_heating_device, t, 0.0) + self._safe_index(
            building.heating_storage.energy_balance, t, 0.0
        )
        if isinstance(building.heating_device, HeatPump):
            heating_kw = self._safe_scalar(
                building.heating_device.get_input_power(heating_demand, temperature, heating=True),
                0.0,
            )
        else:
            heating_kw = self._safe_scalar(building.heating_device.get_input_power(heating_demand), 0.0)

        dhw_demand = self._safe_index(building.energy_from_dhw_device, t, 0.0) + self._safe_index(
            building.dhw_storage.energy_balance, t, 0.0
        )
        if isinstance(building.dhw_device, HeatPump):
            dhw_kw = self._safe_scalar(building.dhw_device.get_input_power(dhw_demand, temperature, heating=True), 0.0)
        else:
            dhw_kw = self._safe_scalar(building.dhw_device.get_input_power(dhw_demand), 0.0)

        non_shiftable_kw = self._safe_index(building.energy_to_non_shiftable_load, t, 0.0)
        solar_kw = self._safe_index(building.solar_generation, t, 0.0)
        washing_kw = sum(self._safe_index(wm.electricity_consumption, t, 0.0) for wm in building.washing_machines or [])

        base_total_kw = cooling_kw + heating_kw + dhw_kw + non_shiftable_kw + solar_kw + washing_kw
        base_phase_kw = self._split_unassigned_power(base_total_kw)

        return float(base_total_kw), base_phase_kw

    def _charger_requested_power_kw(self, charger, action: float) -> float:
        if action is None:
            return 0.0

        action = self._safe_scalar(action, 0.0)
        action = float(np.clip(action, -1.0, 1.0))
        if action > 0.0:
            max_power = self._safe_scalar(getattr(charger, 'max_charging_power', 0.0), 0.0)
            return action * max_power if max_power > 0.0 else 0.0
        if action < 0.0:
            max_power = self._safe_scalar(getattr(charger, 'max_discharging_power', 0.0), 0.0)
            return -abs(action) * max_power if max_power > 0.0 else 0.0
        return 0.0

    def _charger_action_from_power_kw(self, charger, target_power_kw: float) -> float:
        target_power_kw = self._safe_scalar(target_power_kw, 0.0)

        if target_power_kw > 0.0:
            max_power = self._safe_scalar(getattr(charger, 'max_charging_power', 0.0), 0.0)
            min_power = self._safe_scalar(getattr(charger, 'min_charging_power', 0.0), 0.0)
            if max_power <= 0.0:
                return 0.0
            if min_power > 0.0 and target_power_kw < min_power:
                return 0.0
            return float(np.clip(target_power_kw / max_power, 0.0, 1.0))

        if target_power_kw < 0.0:
            max_power = self._safe_scalar(getattr(charger, 'max_discharging_power', 0.0), 0.0)
            min_power = self._safe_scalar(getattr(charger, 'min_discharging_power', 0.0), 0.0)
            requested = abs(target_power_kw)
            if max_power <= 0.0:
                return 0.0
            if min_power > 0.0 and requested < min_power:
                return 0.0
            return float(-np.clip(requested / max_power, 0.0, 1.0))

        return 0.0

    def _storage_requested_power_kw(self, action: Optional[float]) -> float:
        building = self.building
        if action is None:
            return 0.0
        action = self._safe_scalar(action, 0.0)
        action = float(np.clip(action, -1.0, 1.0))
        nominal_power = self._safe_scalar(getattr(building.electrical_storage, 'nominal_power', 0.0), 0.0)
        return action * nominal_power if nominal_power > 0.0 else 0.0

    def _storage_action_from_power_kw(self, target_power_kw: float) -> float:
        building = self.building
        nominal_power = self._safe_scalar(getattr(building.electrical_storage, 'nominal_power', 0.0), 0.0)
        if nominal_power <= 0.0:
            return 0.0
        return float(np.clip(target_power_kw / nominal_power, -1.0, 1.0))

    def _compute_totals(self, base_total_kw: float, base_phase_kw: Mapping[str, float], controls, scales):
        total_kw = float(base_total_kw)
        phase_kw = {phase: float(value) for phase, value in base_phase_kw.items()}

        for control_id, control in controls.items():
            scale = self._safe_scalar(scales.get(control_id, 1.0), 1.0)
            total_kw += control['request_total_kw'] * scale
            for phase_name, value in control['request_phase_kw'].items():
                phase_kw[phase_name] = phase_kw.get(phase_name, 0.0) + (value * scale)

        return total_kw, phase_kw

    def _scale_for_import_scope(self, current_value_kw, limit_kw, controls, scales, component_getter) -> bool:
        limit_kw = self._safe_scalar(limit_kw, np.nan)
        current_value_kw = self._safe_scalar(current_value_kw, 0.0)
        if not np.isfinite(limit_kw):
            return False
        if limit_kw is None or current_value_kw <= limit_kw + 1e-9:
            return False

        relevant = []
        for control_id, control in controls.items():
            component_kw = component_getter(control)
            if component_kw > 0.0 and self._safe_scalar(scales.get(control_id, 0.0), 0.0) > 0.0:
                relevant.append((control_id, component_kw))

        if not relevant:
            return False

        current_relevant_kw = sum(scales[control_id] * component_kw for control_id, component_kw in relevant)
        if current_relevant_kw <= 1e-9:
            return False

        fixed_kw = current_value_kw - current_relevant_kw
        allowed_kw = limit_kw - fixed_kw
        factor = 0.0 if allowed_kw <= 0.0 else min(1.0, allowed_kw / current_relevant_kw)
        if factor >= 1.0 - 1e-9:
            return False

        for control_id, _ in relevant:
            scales[control_id] *= factor

        return True

    def _scale_for_export_scope(self, current_value_kw, limit_kw, controls, scales, component_getter) -> bool:
        limit_kw = self._safe_scalar(limit_kw, np.nan)
        current_value_kw = self._safe_scalar(current_value_kw, 0.0)
        if not np.isfinite(limit_kw):
            return False
        if limit_kw is None:
            return False

        current_export_kw = max(-current_value_kw, 0.0)
        if current_export_kw <= limit_kw + 1e-9:
            return False

        relevant = []
        for control_id, control in controls.items():
            component_kw = component_getter(control)
            if component_kw < 0.0 and self._safe_scalar(scales.get(control_id, 0.0), 0.0) > 0.0:
                relevant.append((control_id, component_kw))

        if not relevant:
            return False

        current_relevant_export_kw = sum(scales[control_id] * abs(component_kw) for control_id, component_kw in relevant)
        if current_relevant_export_kw <= 1e-9:
            return False

        fixed_kw = current_value_kw + current_relevant_export_kw
        allowed_export_kw = limit_kw + fixed_kw
        factor = 0.0 if allowed_export_kw <= 0.0 else min(1.0, allowed_export_kw / current_relevant_export_kw)
        if factor >= 1.0 - 1e-9:
            return False

        for control_id, _ in relevant:
            scales[control_id] *= factor

        return True

    def _apply_legacy_charging_constraints(self, actions: Optional[Mapping[str, float]]) -> Optional[Mapping[str, float]]:
        building = self.building

        if not actions:
            building._set_default_charging_headroom()
            return actions

        positive_requests = {}
        scales = {}
        for charger_id, action in actions.items():
            if action is None or action <= 0.0:
                continue
            charger = building._charger_lookup.get(charger_id)
            if charger is None:
                continue
            max_power = getattr(charger, 'max_charging_power', 0.0) or 0.0
            if max_power <= 0.0:
                continue
            positive_requests[charger_id] = action * max_power
            scales[charger_id] = 1.0

        violation_kw = 0.0

        if positive_requests:
            total_kw = sum(positive_requests.values())
            building_limit = building._building_charger_limit_kw
            building_limit = self._safe_scalar(building_limit, np.nan)
            if np.isfinite(building_limit) and building_limit >= 0.0 and total_kw > building_limit:
                scale = 0.0 if building_limit == 0 else building_limit / total_kw
                for charger_id in scales:
                    scales[charger_id] *= scale
                violation_kw += total_kw - building_limit

            for phase in building._phase_limits:
                limit = phase.get('limit_kw')
                limit = self._safe_scalar(limit, np.nan)
                if not np.isfinite(limit) or limit < 0.0:
                    continue
                chargers = phase.get('chargers', []) or []
                phase_sum = sum(
                    positive_requests.get(charger_id, 0.0) * scales.get(charger_id, 1.0)
                    for charger_id in chargers
                    if charger_id in positive_requests
                )
                if phase_sum > limit:
                    phase_scale = 0.0 if limit == 0 else limit / phase_sum
                    for charger_id in chargers:
                        if charger_id in scales:
                            scales[charger_id] *= phase_scale
                    violation_kw += phase_sum - limit

            scaled_positive_kw = {
                charger_id: positive_requests[charger_id] * scales.get(charger_id, 1.0)
                for charger_id in positive_requests
            }
            used_kw = sum(scaled_positive_kw.values())

            actions = dict(actions)
            for charger_id, action in list(actions.items()):
                if action is None or action <= 0.0:
                    continue
                charger = building._charger_lookup.get(charger_id)
                if charger is None:
                    continue
                max_power = getattr(charger, 'max_charging_power', 0.0) or 0.0
                if max_power <= 0.0:
                    actions[charger_id] = 0.0
                    continue
                target_kw = scaled_positive_kw.get(charger_id, 0.0)
                actions[charger_id] = max(0.0, min(action, target_kw / max_power))

            if getattr(building, '_expose_charging_constraints', False):
                building_limit = self._safe_scalar(building._building_charger_limit_kw, np.nan)
                building_headroom = None if not np.isfinite(building_limit) else building_limit - used_kw
                phase_headroom = {}
                for phase in building._phase_limits:
                    limit = phase.get('limit_kw')
                    limit = self._safe_scalar(limit, np.nan)
                    if not np.isfinite(limit):
                        phase_headroom[phase['name']] = None
                    else:
                        used = sum(scaled_positive_kw.get(charger_id, 0.0) for charger_id in phase.get('chargers', []))
                        phase_headroom[phase['name']] = limit - used

                building._charging_constraints_state = {
                    'building_headroom_kw': building_headroom,
                    'building_export_headroom_kw': None,
                    'phase_headroom_kw': phase_headroom,
                    'phase_export_headroom_kw': {},
                    'total_power_kw': used_kw,
                    'phase_power_kw': {},
                }

            penalty_kwh = self._safe_scalar(violation_kw * (building.seconds_per_time_step / 3600), 0.0)
            building._charging_constraint_penalty_kwh = penalty_kwh
            building._charging_constraint_last_penalty_kwh = penalty_kwh
            phase_power = {}
            if getattr(building, '_electrical_service_enabled', False):
                phase_power = dict((building._charging_constraints_state or {}).get('phase_power_kw') or {})
            building._record_charging_constraint_state(
                violation_kwh=penalty_kwh,
                total_power_kw=float((building._charging_constraints_state or {}).get('total_power_kw', used_kw)),
                phase_power_kw=phase_power,
            )

        else:
            building._set_default_charging_headroom()
            building._record_charging_constraint_state(
                violation_kwh=0.0,
                total_power_kw=0.0,
                phase_power_kw={},
            )

        return actions

    def _apply_electrical_service_constraints(
        self,
        actions: Optional[Mapping[str, float]],
        electrical_storage_action: Optional[float],
    ) -> Tuple[Optional[Mapping[str, float]], Optional[float]]:
        building = self.building
        phase_names = self._current_phase_names()
        base_total_kw, base_phase_kw = self._estimate_non_controllable_base_power()
        base_phase_kw = {phase: base_phase_kw.get(phase, 0.0) for phase in phase_names}

        controls = {}
        adjusted_actions = None if actions is None else dict(actions)

        for charger_id, action in (actions or {}).items():
            charger = building._charger_lookup.get(charger_id)
            if charger is None:
                continue

            request_total_kw = self._charger_requested_power_kw(charger, action)
            if abs(request_total_kw) <= 1e-9:
                continue

            phase_connection = building._charger_phase_map.get(charger_id)
            request_phase_kw = self._split_power_by_connection(request_total_kw, phase_connection)
            controls[charger_id] = {
                'request_total_kw': request_total_kw,
                'request_phase_kw': request_phase_kw,
            }

        storage_control_id = '__electrical_storage__'
        request_storage_kw = self._storage_requested_power_kw(electrical_storage_action)
        if abs(request_storage_kw) > 1e-9:
            request_phase_kw = self._split_power_by_connection(request_storage_kw, building.electrical_storage_phase_connection)
            controls[storage_control_id] = {
                'request_total_kw': request_storage_kw,
                'request_phase_kw': request_phase_kw,
            }

        scales = {control_id: 1.0 for control_id in controls}
        total_limits = building._electrical_service_limits.get('total', {})
        per_phase_limits = building._electrical_service_limits.get('per_phase', {})

        for _ in range(8):
            changed = False
            total_kw, phase_kw = self._compute_totals(base_total_kw, base_phase_kw, controls, scales)

            changed |= self._scale_for_import_scope(
                total_kw,
                total_limits.get('import_kw'),
                controls,
                scales,
                component_getter=lambda c: c['request_total_kw'],
            )
            changed |= self._scale_for_export_scope(
                total_kw,
                total_limits.get('export_kw'),
                controls,
                scales,
                component_getter=lambda c: c['request_total_kw'],
            )

            for phase_name in phase_names:
                phase_limit = per_phase_limits.get(phase_name, {})
                changed |= self._scale_for_import_scope(
                    phase_kw.get(phase_name, 0.0),
                    phase_limit.get('import_kw'),
                    controls,
                    scales,
                    component_getter=lambda c, p=phase_name: c['request_phase_kw'].get(p, 0.0),
                )
                changed |= self._scale_for_export_scope(
                    phase_kw.get(phase_name, 0.0),
                    phase_limit.get('export_kw'),
                    controls,
                    scales,
                    component_getter=lambda c, p=phase_name: c['request_phase_kw'].get(p, 0.0),
                )

            if not changed:
                break

        total_kw, phase_kw = self._compute_totals(base_total_kw, base_phase_kw, controls, scales)
        total_kw = self._safe_scalar(total_kw, 0.0)
        phase_kw = {phase: self._safe_scalar(value, 0.0) for phase, value in phase_kw.items()}

        if adjusted_actions is not None:
            for charger_id in adjusted_actions:
                charger = building._charger_lookup.get(charger_id)
                if charger is None:
                    continue
                control = controls.get(charger_id)
                target_kw = 0.0 if control is None else control['request_total_kw'] * scales.get(charger_id, 1.0)
                adjusted_actions[charger_id] = self._charger_action_from_power_kw(charger, target_kw)

        adjusted_storage_action = electrical_storage_action
        if electrical_storage_action is not None:
            storage_control = controls.get(storage_control_id)
            target_kw = 0.0 if storage_control is None else storage_control['request_total_kw'] * scales.get(storage_control_id, 1.0)
            adjusted_storage_action = self._storage_action_from_power_kw(target_kw)

        violation_kw = 0.0
        import_limit = self._safe_scalar(total_limits.get('import_kw'), np.nan)
        export_limit = self._safe_scalar(total_limits.get('export_kw'), np.nan)
        if np.isfinite(import_limit):
            violation_kw += max(total_kw - import_limit, 0.0)
        if np.isfinite(export_limit):
            violation_kw += max(-total_kw - export_limit, 0.0)

        phase_headroom = {}
        phase_export_headroom = {}
        for phase_name in phase_names:
            phase_total = self._safe_scalar(phase_kw.get(phase_name, 0.0), 0.0)
            phase_limit = per_phase_limits.get(phase_name, {})
            phase_import_limit = self._safe_scalar(phase_limit.get('import_kw'), np.nan)
            phase_export_limit = self._safe_scalar(phase_limit.get('export_kw'), np.nan)

            phase_headroom[phase_name] = None if not np.isfinite(phase_import_limit) else (phase_import_limit - phase_total)
            phase_export_headroom[phase_name] = None if not np.isfinite(phase_export_limit) else (phase_export_limit + phase_total)

            if np.isfinite(phase_import_limit):
                violation_kw += max(phase_total - phase_import_limit, 0.0)
            if np.isfinite(phase_export_limit):
                violation_kw += max(-phase_total - phase_export_limit, 0.0)

        building_headroom = None if not np.isfinite(import_limit) else (import_limit - total_kw)
        building_export_headroom = None if not np.isfinite(export_limit) else (export_limit + total_kw)
        building._charging_constraints_state = {
            'building_headroom_kw': building_headroom,
            'building_export_headroom_kw': building_export_headroom,
            'phase_headroom_kw': phase_headroom,
            'phase_export_headroom_kw': phase_export_headroom,
            'total_power_kw': total_kw,
            'phase_power_kw': phase_kw,
        }

        penalty_kwh = self._safe_scalar(violation_kw * (building.seconds_per_time_step / 3600.0), 0.0)
        building._charging_constraint_penalty_kwh = penalty_kwh
        building._charging_constraint_last_penalty_kwh = penalty_kwh
        building._record_charging_constraint_state(
            violation_kwh=penalty_kwh,
            total_power_kw=float(total_kw),
            phase_power_kw=phase_kw,
        )

        return adjusted_actions, adjusted_storage_action

    def apply_charging_constraints_to_actions(
        self,
        actions: Optional[Mapping[str, float]],
        electrical_storage_action: Optional[float] = None,
    ) -> Tuple[Optional[Mapping[str, float]], Optional[float]]:
        """Apply configured electrical constraints and return adjusted EV/storage actions."""

        building = self.building

        building._charging_constraint_penalty_kwh = 0.0
        building._charging_constraint_last_penalty_kwh = 0.0

        if not building._charging_constraints_enabled:
            return actions, electrical_storage_action

        if getattr(building, '_electrical_service_enabled', False):
            return self._apply_electrical_service_constraints(actions, electrical_storage_action)

        adjusted_actions = self._apply_legacy_charging_constraints(actions)
        return adjusted_actions, electrical_storage_action



from pathlib import Path
from typing import List, Mapping, Tuple, Union
import numpy as np
from sklearn.tree import DecisionTreeClassifier
from citylearn.base import Environment
from citylearn.data import LogisticRegressionOccupantParameters
from citylearn.utilities import FileHandler

class Occupant(Environment):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def predict(self) -> float:
        delta = 0.0

        return delta

class LogisticRegressionOccupant(Occupant):
    def __init__(
            self, setpoint_increase_model_filepath: Union[Path, str], setpoint_decrease_model_filepath: Union[Path, str], 
            delta_output_map: Mapping[int, float], parameters: LogisticRegressionOccupantParameters, **kwargs
        ):
        super().__init__(**kwargs)
        self.__setpoint_increase_model: DecisionTreeClassifier = None
        self.__setpoint_decrease_model: DecisionTreeClassifier = None
        self.__probabilities = None
        self.setpoint_increase_model_filepath = setpoint_increase_model_filepath
        self.setpoint_decrease_model_filepath = setpoint_decrease_model_filepath
        self.delta_output_map = delta_output_map
        self.parameters = parameters

    @property
    def probabilities(self) -> Mapping[str, float]:
        return self.__probabilities

    @property
    def setpoint_increase_model_filepath(self) -> Union[Path, str]:
        return self.__setpoint_increase_model_filepath
    
    @property
    def setpoint_decrease_model_filepath(self) -> Union[Path, str]:
        return self.__setpoint_decrease_model_filepath
    
    @property
    def delta_output_map(self) -> Mapping[int, float]:
        return self.__delta_output_map
    
    @setpoint_increase_model_filepath.setter
    def setpoint_increase_model_filepath(self, value: Union[Path, str]):
        self.__setpoint_increase_model_filepath = value
        self.__setpoint_increase_model = FileHandler.read_pickle(self.setpoint_increase_model_filepath)

    @setpoint_decrease_model_filepath.setter
    def setpoint_decrease_model_filepath(self, value: Union[Path, str]):
        self.__setpoint_decrease_model_filepath = value
        self.__setpoint_decrease_model = FileHandler.read_pickle(self.setpoint_decrease_model_filepath)

    @delta_output_map.setter
    def delta_output_map(self, value: Mapping[Union[str, int], float]):
        self.__delta_output_map = {int(k): v for k, v in value.items()}

    def predict(self, x: Tuple[float, List[List[float]]]) -> float:
        delta = super().predict()
        response = None
        interaction_input, delta_input = x
        interaction_probability = lambda  a, b, x_ : 1/(1 + np.exp(-(a + b*x_)))
        increase_setpoint_probability = interaction_probability(self.parameters.a_increase[self.time_step], self.parameters.b_increase[self.time_step], interaction_input)
        decrease_setpoint_probability = interaction_probability(self.parameters.a_decrease[self.time_step], self.parameters.b_decrease[self.time_step], interaction_input)
        random_seed = max(self.random_seed, 1) + self.time_step
        nprs = np.random.RandomState(random_seed)
        random_probability = nprs.uniform()
        self.__probabilities['increase_setpoint'][self.time_step] = increase_setpoint_probability
        self.__probabilities['decrease_setpoint'][self.time_step] = decrease_setpoint_probability
        self.__probabilities['random'][self.time_step] = random_probability
        
        if (increase_setpoint_probability < random_probability and decrease_setpoint_probability < random_probability) \
                or (increase_setpoint_probability >= random_probability and decrease_setpoint_probability >= random_probability):
            pass

        elif increase_setpoint_probability >= random_probability:
            response = self.__setpoint_increase_model.predict(delta_input)
            delta = self.delta_output_map[response[0]]

        elif decrease_setpoint_probability >= random_probability:
            response = self.__setpoint_decrease_model.predict(delta_input)
            delta = -self.delta_output_map[response[0]]

        else:
            pass

        return delta
    
    def reset(self):
        super().reset()
        self.__probabilities = {
            'increase_setpoint': np.zeros(self.episode_tracker.episode_time_steps, dtype='float32'),
            'decrease_setpoint': np.zeros(self.episode_tracker.episode_time_steps, dtype='float32'),
            'random': np.zeros(self.episode_tracker.episode_time_steps, dtype='float32'),
        }


from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


@pytest.fixture
def env():
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        episode_time_steps=96,
        random_seed=0,
    )
    env.reset()
    try:
        yield env
    finally:
        env.close()


def _zero_actions(env: CityLearnEnv):
    return [np.zeros(space.shape, dtype="float32") for space in env.action_space]


def _ev_action_index(env: CityLearnEnv, building, charger_id: str) -> int:
    offset = 0
    for b in env.buildings:
        if b is building:
            break
        offset += b.action_space.shape[0]
    else:  # pragma: no cover - defensive guard
        raise AssertionError("Building not found while resolving action index.")

    action_key = f"electric_vehicle_storage_{charger_id}"
    local_index = building.active_actions.index(action_key)
    return offset + local_index


def _get_charger(building, charger_id: str):
    for charger in building.electric_vehicle_chargers or []:
        if charger.charger_id == charger_id:
            return charger

    raise AssertionError(f"Charger {charger_id} not found in building {building.name}")


def test_soc_increases_when_connected_ev_is_charged(env: CityLearnEnv):
    while env.time_step < env.episode_time_steps - 1:
        t = env.time_step

        for building in env.buildings:
            if not building.electric_vehicle_chargers:
                continue

            for charger in building.electric_vehicle_chargers:
                ev = charger.connected_electric_vehicle
                if ev is None:
                    continue

                sim = charger.charger_simulation
                if t + 1 >= len(sim.electric_vehicle_charger_state):
                    continue

                connected_now = (
                    isinstance(sim.electric_vehicle_id[t], str)
                    and sim.electric_vehicle_id[t] == ev.name
                    and sim.electric_vehicle_charger_state[t] == 1
                )
                connected_next = sim.electric_vehicle_charger_state[t + 1] == 1

                if connected_now and connected_next:
                    initial_soc = float(ev.battery.soc[t])
                    actions = _zero_actions(env)
                    actions[0][_ev_action_index(env, building, charger.charger_id)] = 1.0
                    env.step(actions)

                    charger_after = _get_charger(building, charger.charger_id)
                    ev_after = charger_after.connected_electric_vehicle
                    assert ev_after is not None

                    new_index = max(env.time_step - 1, 0)
                    new_soc = float(ev_after.battery.soc[new_index])
                    assert new_soc > initial_soc + 1e-3
                    return

        if env.time_step >= env.episode_time_steps - 1:
            break
        env.step(_zero_actions(env))

    pytest.fail("No EV remained connected long enough for SOC charging test.")


def test_soc_matches_dataset_on_arrival(env: CityLearnEnv):
    while env.time_step < env.episode_time_steps - 1:
        t = env.time_step

        for building in env.buildings:
            if not building.electric_vehicle_chargers:
                continue

            for charger in building.electric_vehicle_chargers:
                sim = charger.charger_simulation
                if t + 1 >= len(sim.electric_vehicle_charger_state):
                    continue

                curr_state = sim.electric_vehicle_charger_state[t]
                next_state = sim.electric_vehicle_charger_state[t + 1]
                next_id = sim.electric_vehicle_id[t + 1] if t + 1 < len(sim.electric_vehicle_id) else ""

                if not isinstance(next_id, str) or not next_id.strip():
                    continue

                if curr_state == 1 or next_state != 1:
                    continue

                if curr_state == 2:
                    if t >= len(sim.electric_vehicle_estimated_soc_arrival):
                        continue
                    expected_soc = float(sim.electric_vehicle_estimated_soc_arrival[t])
                else:
                    if t + 1 >= len(sim.electric_vehicle_estimated_soc_arrival):
                        continue
                    expected_soc = float(sim.electric_vehicle_estimated_soc_arrival[t + 1])

                if not (0.0 <= expected_soc <= 1.0):
                    continue

                env.step(_zero_actions(env))

                charger_after = _get_charger(building, charger.charger_id)
                ev_after = charger_after.connected_electric_vehicle
                if ev_after is None or ev_after.name != next_id:
                    continue

                new_soc = float(ev_after.battery.soc[env.time_step])
                assert pytest.approx(expected_soc, rel=0, abs=1e-5) == new_soc
                return

        if env.time_step >= env.episode_time_steps - 1:
            break
        env.step(_zero_actions(env))

    pytest.fail("No EV arrival transition observed during simulation.")



import numpy as np
import pandas as pd
import pytest

from glob import glob

pytest.importorskip("gymnasium")

from citylearn.citylearn import CityLearnEnv


SCHEMA_PATH = "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


def _zero_actions(env: CityLearnEnv):
    return [np.zeros(space.shape, dtype="float32") for space in env.action_space]


def _find_transition(from_state: int):
    for csv_path in glob("data/datasets/citylearn_challenge_2022_phase_all_plus_evs/charger_*_*.csv"):
        df = pd.read_csv(csv_path)
        for idx in range(len(df) - 1):
            if (
                df.loc[idx, "electric_vehicle_charger_state"] == from_state
                and df.loc[idx + 1, "electric_vehicle_charger_state"] == 1
            ):
                return csv_path, idx

    raise AssertionError(f"No {from_state}->1 transition found in dataset.")


@pytest.mark.parametrize("from_state", [2, 3])
def test_ev_soc_matches_dataset_on_arrival(from_state: int):
    csv_path, transition_index = _find_transition(from_state)
    charger_id = csv_path.split("/")[-1].replace(".csv", "")

    env = CityLearnEnv(SCHEMA_PATH, central_agent=True, random_seed=0)
    env.reset()

    # Advance the environment to the timestep of interest.
    for _ in range(transition_index + 1):
        env.step(_zero_actions(env))

    # Identify the EV connected to the selected charger.
    target_ev_id = None

    for building in env.buildings:
        for charger in building.electric_vehicle_chargers or []:
            if charger.charger_id == charger_id:
                sim = charger.charger_simulation
                target_ev_id = sim.electric_vehicle_id[transition_index + 1]
                break
        if target_ev_id:
            break

    assert target_ev_id, "Expected EV id for charger transition not found."

    ev = next(ev for ev in env.electric_vehicles if ev.name == target_ev_id)

    step = env.time_step
    prev_state = sim.electric_vehicle_charger_state[step - 1] if step > 0 else float("nan")
    prev_ev_id = sim.electric_vehicle_id[step - 1] if step > 0 else None

    candidate_index = None
    if prev_state in (2, 3):
        candidate_index = step - 1
    elif 0 <= step < len(sim.electric_vehicle_estimated_soc_arrival):
        candidate_index = step

    estimated_soc = None
    if candidate_index is not None and 0 <= candidate_index < len(sim.electric_vehicle_estimated_soc_arrival):
        candidate_value = sim.electric_vehicle_estimated_soc_arrival[candidate_index]
        if isinstance(candidate_value, (float, np.floating)) and not np.isnan(candidate_value) and candidate_value >= 0:
            estimated_soc = float(candidate_value)

    if estimated_soc is None:
        fallback_index = step if step < len(sim.electric_vehicle_required_soc_departure) else step - 1
        fallback_value = sim.electric_vehicle_required_soc_departure[fallback_index]
        assert (
            isinstance(fallback_value, (float, np.floating))
            and not np.isnan(fallback_value)
            and fallback_value >= 0
        ), "Expected fallback SOC from required departure."
        expected_soc = float(fallback_value)
    else:
        expected_soc = estimated_soc

    assert pytest.approx(expected_soc, abs=1e-6) == float(ev.battery.soc[env.time_step])


def test_ev_kpi_evaluation_with_evs_and_chargers():
    env = CityLearnEnv(SCHEMA_PATH, central_agent=True, random_seed=0)
    env.reset()

    for _ in range(10):
        env.step(_zero_actions(env))

    df = env.evaluate()

    district_values = df[df["level"] == "district"]["value"]
    assert district_values.notna().any(), "District-level KPI values should contain finite entries when EVs are present."


def test_ev_current_soc_overrides_arrival_estimate_when_present():
    csv_path, transition_index = _find_transition(2)
    charger_id = csv_path.split("/")[-1].replace(".csv", "")

    env = CityLearnEnv(SCHEMA_PATH, central_agent=True, random_seed=0)
    env.reset()

    target_charger = None
    for building in env.buildings:
        for charger in building.electric_vehicle_chargers or []:
            if charger.charger_id == charger_id:
                target_charger = charger
                break
        if target_charger is not None:
            break

    assert target_charger is not None, "Expected charger for transition was not found."
    sim = target_charger.charger_simulation
    forced_soc = 0.42
    current_soc = np.full(len(sim.electric_vehicle_charger_state), -0.1, dtype="float32")
    current_soc[transition_index + 1] = forced_soc
    sim.electric_vehicle_current_soc = current_soc

    for _ in range(transition_index + 1):
        env.step(_zero_actions(env))

    connected_ev = target_charger.connected_electric_vehicle
    assert connected_ev is not None, "Expected EV to be connected at the transition step."
    assert float(connected_ev.battery.soc[env.time_step]) == pytest.approx(forced_soc, abs=1e-6)



from typing import Any, List, Mapping, Tuple
import numpy as np
import pandas as pd
import torch
from torch.utils.data import TensorDataset, DataLoader
from citylearn.building import Building
from citylearn.data import get_settings
from citylearn.preprocessing import Normalize, PeriodicNormalization

def preprocess_df(config:  Mapping[str, Any], df: pd.DataFrame, train_references: List[int] = None, validation_references: List[int] = None, test_references: List[int] = None) -> Mapping[str, Any]:
    ideal_reference = 0
    # # including in training makes model worse so will exclude for now. 
    # # Nevertheless, prediction for free-float is still decent despite not including it in the training.
    # free_float_reference = 1 
    partial_references = [2, 3, 4, 5]
    train_references = partial_references[:2] if train_references is None else train_references
    validation_references = [partial_references[2]] if validation_references is None else validation_references
    test_references = [partial_references[3]] if test_references is None else test_references
    
    # observation names
    observation_names = get_settings()['schema']['template']['buildings']['Building_1']['dynamics']['attributes']['input_observation_names']
    target = observation_names[-1]
    periodic_observations = Building.get_periodic_observation_metadata()
    
    # periodic normalization
    for k, v in periodic_observations.items():
        result = df[k]*PeriodicNormalization(x_max=v[-1])
        result = pd.DataFrame(result.tolist(), index=result.index)
        df[f'{k}_sin'] = result[0].tolist()
        df[f'{k}_cos'] = result[1].tolist()

    # set min-max normalization limits
    normalization_minimum = df[observation_names].min().values.tolist()
    normalization_maximum = df[observation_names].max().values.tolist()

    # min-max normalization
    for c in observation_names:
        df[c] = df[c]*Normalize(df[c].min(), df[c].max())

    # check to make share there is 1-year worth of data
    months = list(sorted(df['month'].unique()))
    assert len(months) == 12, f'Expected 12 months, got {len(months)} months.'
    
    # training data are ideal load data for every three month step beginning from January
    # and free-float load and 2 partial load datasets for entire year
    train_df = df[
        ((df['reference']==ideal_reference) & (df['month'].isin([months[i] for i in range(0, len(months), 3)])))
        | (df['reference'].isin(train_references))
    ][observation_names].copy()
    X_train, y_train = sliding_windows(train_df.to_numpy(), config['lb'], 1)
    train_df, train_loader = dataset_dataloader(X_train, y_train, config['batch_size'])
    
    # validation data are ideal load data for every three month step beginning from February
    # and free-float load and 1 partial load dataset for entire year
    validation_df = df[
        ((df['reference']==ideal_reference) & (df['month'].isin([months[i] for i in range(1, len(months), 3)])))
        | (df['reference'].isin(validation_references))
    ][observation_names].copy()
    X_val, y_val = sliding_windows(validation_df.to_numpy(), config['lb'], 1)
    val_df, val_loader = dataset_dataloader(X_val, y_val, config['batch_size'])

    # test data are ideal load data for every three month step beginning from March
    # and free-float load and 1 partial load dataset for entire year
    test_df = df[
        ((df['reference']==ideal_reference) & (df['month'].isin([months[i] for i in range(2, len(months), 3)])))
        | (df['reference'].isin(test_references))
    ].copy()
    test_df_by_season = test_df.copy()
    test_df = test_df[observation_names].copy()
    X_test, y_test = sliding_windows(test_df.to_numpy(), config['lb'], 1)
    test_df, test_loader = dataset_dataloader(X_test, y_test, config['batch_size'])
    
    # seasonal test data is same as test data but in 3-month sequences
    test_df_by_season = [
        test_df_by_season[test_df_by_season['month'].isin(months[i:i+3])][observation_names] 
        for i in range(0, len(months), 3)
    ]
    test_loader_by_season = [dataset_dataloader(
        *sliding_windows(df.to_numpy(), config['lb'], 1), 
        config['batch_size']
    )[1] for df in test_df_by_season]
    
    return {
        'temp_limits': {
            'min': normalization_minimum[observation_names.index(target)],
            'max': normalization_maximum[observation_names.index(target)]
        },
        'loaders': {
            'train': train_loader,
            'val': val_loader,
            'test': test_loader,
            'test_by_season':  test_loader_by_season
        },
        'train': {
            'X': X_train,
            'y': y_train
        },
        'val': {
            'X': X_val,
            'y': y_val
        },
        'test': {
            'X': X_test,
            'y': y_test
        },
        'observation_metadata': {
            'input_observation_names': observation_names,
            'input_normalization_minimum': normalization_minimum,
            'input_normalization_maximum': normalization_maximum, 
        }
    }

def dataset_dataloader(x: np.ndarray, y: np.ndarray, batch_size: int, shuffle: bool = None, drop_last: bool = None) -> Tuple[TensorDataset, DataLoader]:
    shuffle = True if shuffle is None else shuffle
    drop_last = True if drop_last is None else drop_last
    tensor = TensorDataset(torch.from_numpy(x.astype(np.float32)), torch.from_numpy(y.astype(np.float32)))
    loader = DataLoader(tensor, shuffle=shuffle, batch_size=batch_size, drop_last=drop_last)
    
    return tensor, loader

def sliding_windows(data: np.ndarray, seq_length: int, output_len: int):
    """
    Check that the variable to be predicted is the last column of the dataframe
    :param data: dataframe
    :param seq_length: lookback
    :param output_len: how many timetep ahead will be predicted
    :return: x = matrix [number of timestep - lookback, lookback, number of input variables];
             y = matrix [number of timestep - lookback, number of output variables]
    """
    
    x = []
    y = []

    for i in range(len(data) - seq_length):
        _x = data[(i+1):(i + 1 + seq_length), :-1]
        T_lag = data[i:(i+seq_length), -1]
        _y = data[(i + seq_length):(i + seq_length + output_len), -1]  # If you want to predict more than one timestamp
        _x = np.column_stack([_x, T_lag])
        x.append(_x)
        y.append(_y)

    return np.array(x), np.array(y)


import os
from pathlib import Path
from typing import Any, List, Mapping, Union
from doe_xstock.simulate import EndUseLoadProfilesEnergyPlusSimulator
from eppy.modeleditor import IDF
import numpy as np
import pandas as pd

class EndUseLoadProfilesEnergyPlusPartialLoadSimulator(EndUseLoadProfilesEnergyPlusSimulator):
    QUERIES_DIRECTORY = os.path.join(os.path.join(*Path(os.path.dirname(os.path.abspath(__file__))).parts[0:-1]), 'misc', 'queries')

    def __init__(
        self, ideal_loads_simulator: EndUseLoadProfilesEnergyPlusSimulator, allow_multi_load_time_step: bool = None, multiplier_minimum: float = None,
        multiplier_maximum: float = None, multiplier_probability: float = None, output_variables: List[str] = None, output_meters: List[str] = None, 
        simulation_id: str = None, output_directory: Union[Path, str] = None, random_seed: int = None,
    ):
        self.ideal_loads_simulator = ideal_loads_simulator
        super().__init__(
            self.ideal_loads_simulator.version, 
            self.ideal_loads_simulator.idd_filepath, 
            self.ideal_loads_simulator.idf, 
            self.ideal_loads_simulator.epw, 
            schedules_filepath=self.ideal_loads_simulator.schedules_filepath, 
            number_of_time_steps_per_hour=self.ideal_loads_simulator.number_of_time_steps_per_hour, 
            output_variables=output_variables if output_variables is not None else self.ideal_loads_simulator.output_variables, 
            output_meters=output_meters if output_meters is not None else self.ideal_loads_simulator.output_meters, 
            osm=False, 
            ideal_loads=False, 
            edit_ems=False, 
            simulation_id=simulation_id,
            output_directory=output_directory,
            cache=self.ideal_loads_simulator.cache,
            idf_preprocessing_customization_function=self.ideal_loads_simulator.idf_preprocessing_customization_function

        )
        self.allow_multi_load_time_step = allow_multi_load_time_step
        self.multiplier_minimum = multiplier_minimum
        self.multiplier_maximum = multiplier_maximum
        self.multiplier_probability = multiplier_probability
        self.random_seed = random_seed
    
    @property
    def ideal_loads_simulator(self) -> EndUseLoadProfilesEnergyPlusSimulator:
        return self.__ideal_loads_simulator
    
    @property
    def allow_multi_load_time_step(self) -> bool:
        return self.__allow_multi_load_time_step
    
    @property
    def multiplier_minimum(self) -> float:
        return self.__multiplier_minimum
    
    @property
    def multiplier_maximum(self) -> float:
        return self.__multiplier_maximum
    
    @property
    def multiplier_probability(self) -> float:
        return self.__multiplier_probability
    
    @property
    def random_seed(self) -> int:
        return self.__random_seed
    
    @ideal_loads_simulator.setter
    def ideal_loads_simulator(self, value: EndUseLoadProfilesEnergyPlusSimulator):
        self.__ideal_loads_simulator = value

    @allow_multi_load_time_step.setter
    def allow_multi_load_time_step(self, value: bool):
        self.__allow_multi_load_time_step = False if value is None else value

    @multiplier_minimum.setter
    def multiplier_minimum(self, value: float):
        self.__multiplier_minimum = 0.3 if value is None else value

    @multiplier_maximum.setter
    def multiplier_maximum(self, value: float):
        self.__multiplier_maximum = 1.7 if value is None else value

    @multiplier_probability.setter
    def multiplier_probability(self, value: float):
        value = 0.6 if value is None else value
        assert 0 <= value <= 1.0, 'multiplier_probability must be > 0 and <= 1.'
        self.__multiplier_probability = value

    @random_seed.setter
    def random_seed(self, value: int):
        self.__random_seed = 0 if value is None else value

    @classmethod
    def multi_simulate(cls, simulators: list, max_workers=None):
        super().multi_simulate(simulators, max_workers=max_workers)

        for s in simulators:
            s.insert_zone_metadata(s)

    def simulate(self, **run_kwargs):
        super().simulate(**run_kwargs)
        self.insert_zone_metadata(self)

    def preprocess_idf_for_simulation(self) -> IDF:
        idf = super().preprocess_idf_for_simulation()
        idf = self.remove_ideal_loads_air_system(idf)
        idf = self.add_other_equipment(idf)

        return idf
    
    def add_other_equipment(self, idf: IDF) -> IDF:
        # schedule type limit object
        schedule_type_limit_name = 'other equipment hvac power'
        obj = idf.newidfobject('ScheduleTypeLimits')
        obj.Name = schedule_type_limit_name
        obj.Lower_Limit_Value = ''
        obj.Upper_Limit_Value = ''
        obj.Numeric_Type = 'Continuous'
        obj.Unit_Type = 'Dimensionless'

        # generate stochastic thermal load
        ideal_loads = self.get_ideal_loads()
        timesteps = ideal_loads['timestep'].max()
        multipliers = self.get_multipliers(ideal_loads.shape[0])
        ideal_loads['cooling_load'] *= -multipliers
        ideal_loads['heating_load'] *= multipliers
        loads_filepath = os.path.join(self.output_directory, f'{self.simulation_id}-partial_load.csv')
        ideal_loads = ideal_loads.sort_values(['zone_name', 'timestep'])
        loads = ['cooling_load', 'heating_load']
        ideal_loads[loads].to_csv(loads_filepath, index=False)
        
        for i, (zone_name, _) in enumerate(ideal_loads.groupby('zone_name')):
            for j, load in enumerate(loads):
                # put schedule obj
                obj = idf.newidfobject('Schedule:File')
                schedule_object_name = f'{zone_name} partial {load}'
                obj.Name = schedule_object_name
                obj.Schedule_Type_Limits_Name = schedule_type_limit_name
                obj.File_Name = loads_filepath
                obj.Column_Number = j + 1
                obj.Rows_to_Skip_at_Top = 1 + i*timesteps
                obj.Number_of_Hours_of_Data = 8760

                if self.number_of_time_steps_per_hour is None:
                    timestep_obj = self.ideal_loads_simulator.get_idf_object().idfobjects['Timestep'][0]
                    number_of_time_steps_per_hour = timestep_obj.Number_of_Timesteps_per_Hour

                else:
                    number_of_time_steps_per_hour = self.number_of_time_steps_per_hour

                obj.Minutes_per_Item = int(60/number_of_time_steps_per_hour)

                # put other equipment
                obj = idf.newidfobject('OtherEquipment')
                obj.Name = f'{zone_name} partial {load}'
                obj.Fuel_Type = 'None'
                obj.Zone_or_ZoneList_or_Space_or_SpaceList_Name = zone_name
                obj.Schedule_Name = schedule_object_name
                obj.Design_Level_Calculation_Method = 'EquipmentLevel'
                obj.Design_Level = 1.0 # already in Watts from query
                obj.Fraction_Latent = 0.0
                obj.Fraction_Radiant = 0.0
                obj.Fraction_Lost = 0.0
                obj.EndUse_Subcategory = f'partial {load}'

        return idf
    
    def get_multipliers(self, size: int) -> np.ndarray:
        nprs = np.random.RandomState(self.random_seed)
        data = nprs.uniform(self.multiplier_minimum, self.multiplier_maximum, size)
        data[nprs.random(size) > self.multiplier_probability] = 1.0
        
        return data
    
    def get_ideal_loads(self):
        self.insert_zone_metadata(self.ideal_loads_simulator)
        zones = self.get_zone_conditioning_metadata()
        cooled_zone_names = [f'\'{k}\'' for k, v in zones.items() if v['is_cooled']==1]
        heated_zone_names = [f'\'{k}\'' for k, v in zones.items() if v['is_heated']==1]
        cooled_zone_names = ', '.join(cooled_zone_names)
        heated_zone_names = ', '.join(heated_zone_names)

        query_filepath = os.path.join(self.QUERIES_DIRECTORY, 'select_ideal_loads.sql')
        data = self.ideal_loads_simulator.get_output_database().query_table_from_file(query_filepath, replace={
            '<cooled_zone_names>': cooled_zone_names, 
            '<heated_zone_names>': heated_zone_names, 
        })
        data = data.pivot(index=['timestep', 'zone_index', 'zone_name'], columns='load', values='value')
        data = data.fillna(0.0)
        data = data.reset_index()

        if not self.allow_multi_load_time_step:
            data.loc[data['cooling_load'] > data['heating_load'], 'heating_load'] = 0.0
            data.loc[data['heating_load'] > data['cooling_load'], 'cooling_load'] = 0.0
        
        else:
            pass

        return data
    
    def insert_zone_metadata(self, simulator: EndUseLoadProfilesEnergyPlusSimulator):
        zones = self.get_zone_conditioning_metadata()
        data = pd.DataFrame([v for _, v in zones.items()])
        query_filepath = os.path.join(self.QUERIES_DIRECTORY, 'create_zone_metadata.sql')
        simulator.get_output_database().execute_sql_from_file(query_filepath)
        simulator.get_output_database().insert('zone_metadata', data.columns.tolist(), data.values)

    def get_zone_conditioning_metadata(self) -> Mapping[str, Mapping[str, Any]]:
        query_filepath = os.path.join(self.QUERIES_DIRECTORY, 'select_zone_conditioning_metadata.sql')
        data = self.ideal_loads_simulator.get_output_database().query_table_from_file(query_filepath)
        zones = {z['zone_name']: z for z in data.to_dict('records')}

        return zones

    def remove_ideal_loads_air_system(self, idf: IDF) -> IDF:
        idf.idfobjects['HVACTemplate:Zone:IdealLoadsAirSystem'] = []
        obj_names = [
            'ZoneControl:Thermostat', 'ZoneControl:Humidistat', 'ZoneControl:Thermostat:ThermalComfort', 'ThermostatSetpoint:DualSetpoint', 
            'ZoneControl:Thermostat:OperativeTemperature', 'ZoneControl:Thermostat:TemperatureAndHumidity', 'ZoneControl:Thermostat:StagedDualSetpoint'
        ]

        for name in obj_names:
            idf.idfobjects[name] = []

        # turn off sizing
        for obj in idf.idfobjects['SimulationControl']:
            obj.Do_Zone_Sizing_Calculation = 'No'
            obj.Do_System_Sizing_Calculation = 'No'
            obj.Do_Plant_Sizing_Calculation = 'No'
            obj.Run_Simulation_for_Sizing_Periods = 'No'

        return idf


from pathlib import Path

import numpy as np
import pytest

pytest.importorskip("gymnasium")

from citylearn.agents.rbc import BasicElectricVehicleRBC_ReferenceController as Agent
from citylearn.citylearn import CityLearnEnv


SCHEMA = Path(__file__).resolve().parents[1] / "data/datasets/citylearn_challenge_2022_phase_all_plus_evs/schema.json"


def _make_zero_actions(env: CityLearnEnv):
    return [np.zeros(space.shape, dtype="float32") for space in env.action_space]


def test_basic_ev_rbc_completes_episode(tmp_path):
    render_dir = tmp_path / "render_evs"
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        render_mode="during",
        render_directory=render_dir,
        episode_time_steps=24,
        random_seed=0,
    )

    try:
        controller = Agent(env)
        controller.learn(episodes=1, logging_level=1)

        kpis = controller.env.evaluate()
        assert not kpis.empty

        # Ensure rendering pipeline produced output when enabled.
        assert render_dir.exists()
    finally:
        env.close()


def test_zero_action_rollout_keeps_kpis_finite():
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        episode_time_steps=12,
        random_seed=0,
    )

    try:
        env.reset()
        while not env.terminated:
            env.step(_make_zero_actions(env))

        kpis = env.evaluate()
        finite = kpis["value"].dropna()
        assert not finite.empty
        assert np.isfinite(finite).all()
    finally:
        env.close()


def test_rbc_sets_ev_actions_positive():
    env = CityLearnEnv(
        str(SCHEMA),
        central_agent=True,
        episode_time_steps=4,
        random_seed=0,
    )

    try:
        controller = Agent(env)
        observations, _ = env.reset()
        actions = controller.predict(observations, deterministic=True)

        ev_indices = [
            idx
            for idx, name in enumerate(env.action_names[0])
            if name.startswith("electric_vehicle_storage_")
        ]
        assert ev_indices, "Expected EV storage actions in action space."

        for idx in ev_indices:
            assert actions[0][idx] > 0.0
    finally:
        env.close()



from __future__ import annotations

from pathlib import Path
from typing import List, Union
try:
    import torch
    import torch.nn
except ImportError:  # pragma: no cover - optional dependency for LSTM dynamics only
    torch = None


class _TorchNNPlaceholder:
    class Module:
        pass

class Dynamics:
    """Base building dynamics model."""

    def __init__(self):
        pass

    def reset(self):
        pass

class LSTMDynamics(Dynamics, (torch.nn if torch is not None else _TorchNNPlaceholder).Module):
    """LSTM building dynamics model that predicts indoor temperature based on partial cooling/heating load and other weather variables.
    
    Parameters
    ----------
    filepath: Union[Path, str]
        Path to model state dictionary.
    input_observation_names: List[str]
        List of maximum values used for input observation min-max normalization.
    input_normalization_minimum: List[float]
        List of minumum values used for input observation min-max normalization.
    input_normalization_maximum: List[float]
        List of maximum values used for input observation min-max normalization.
    hidden_size: int
        The number of neurons in hidden layer.
    num_layers: int
        Number of hidden layers.
    lookback: int
        Number of samples used for prediction.
    input_size: int, optional
        Number of variables used for prediction. This may not equal `input_observation_names`
        e.g. cooling and heating demand may be included in `input_observation_names` but only
        one of two may be used for the actual prediction depending on building needs.
        The default is to set set `input_size` to the length of `input_observation_names`.
    dropout: float, default: 0.0
        Probability of excluding input and recurrent connections to LSTM units from activation 
        and weight updates while training a network. This has the effect of reducing overfitting 
        and improving model performance.
    """

    def __init__(
            self, filepath: Union[Path, str], input_observation_names: List[str], input_normalization_minimum: List[float], 
            input_normalization_maximum: List[float], hidden_size: int, num_layers: int, lookback: int, input_size: int = None,
            dropout: float = None
    ):
        if torch is None:
            raise ImportError('torch is required to use LSTMDynamics.')

        Dynamics.__init__(self)
        torch.nn.Module.__init__(self)
        assert len(input_observation_names) == len(input_normalization_minimum) == len(input_normalization_maximum),\
            'input_observation_names, input_normalization_minimum and input_normalization_maximum must have the same length.'
        self.input_observation_names = input_observation_names
        self.input_normalization_minimum = input_normalization_minimum
        self.input_normalization_maximum = input_normalization_maximum
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.lookback = lookback
        self.filepath = filepath
        self.l_lstm = self.set_lstm()
        self.l_linear = self.set_linear()
        self._hidden_state = None
        self._model_input = None
        self.dropout = torch.nn.Dropout(dropout if dropout is not None else 0.0)

    @property
    def input_size(self) -> int:
        return self.__input_size
    
    @input_size.setter
    def input_size(self, value: int):
        self.__input_size = len(self.input_observation_names) if value is None else value
    
    def set_lstm(self) -> torch.nn.LSTM:
        """Initialize LSTM model."""

        return torch.nn.LSTM(
            input_size=self.input_size,
            hidden_size=self.hidden_size,
            num_layers=self.num_layers,
            batch_first=True,
        )

    def set_linear(self) -> torch.nn.Linear:
        """Initialize linear transformer."""

        return torch.nn.Linear(
            in_features=self.hidden_size,
            out_features=1
        )
    
    def forward(self, x, h):
        """Predict indoor dry bulb temperature."""

        lstm_out, h = self.l_lstm(x, h)
        lstm_out = self.dropout(lstm_out)
        out = lstm_out[:, -1, :]
        out_linear_transf = self.l_linear(out)
        return out_linear_transf, h

    def init_hidden(self, batch_size: int):
        """Initialize hidden states."""

        hidden_state = torch.zeros(self.num_layers, batch_size, self.hidden_size)
        cell_state = torch.zeros(self.num_layers, batch_size, self.hidden_size)
        hidden = (hidden_state, cell_state)
        
        return hidden

    def reset(self):
        """Loads dynamic model state dict, and initializes hidden states and model input."""

        super().reset()

        try:
            self.load_state_dict(torch.load(self.filepath)['model_state_dict'])
        
        except RuntimeError:
            self.load_state_dict(torch.load(self.filepath, map_location=torch.device('cpu'))['model_state_dict'])
        
        except:
            self.load_state_dict(torch.load(self.filepath))

        self._hidden_state = self.init_hidden(1)
        self._model_input = [[None]*(self.lookback + 1) for _ in self.input_observation_names]

    def terminate(self):
        return



import concurrent.futures
import random
from typing import List, Tuple
from doe_xstock.end_use_load_profiles import EndUseLoadProfiles
import math
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.metrics import cluster as cluster_metrics
from sklearn.preprocessing import MinMaxScaler
from citylearn.base import Environment

class Clustering:
    __MINIMUM_BUILDING_COUNT = 3

    def __init__(self, end_use_load_profiles: EndUseLoadProfiles, bldg_ids: List[int], maximum_clusters: int = None, sum_of_squares_error_minimum_percent_change: float = None, random_seed: int = None) -> None:
        self.end_use_load_profiles = end_use_load_profiles
        self.bldg_ids = bldg_ids
        self.maximum_clusters = maximum_clusters
        self.sum_of_squares_error_minimum_percent_change = sum_of_squares_error_minimum_percent_change
        self.random_seed = random_seed

    @property
    def maximum_clusters(self) -> int:
        return self.__maximum_clusters
    
    @property
    def bldg_ids(self) -> List[int]:
        return self.__bldg_ids
    
    @property
    def sum_of_squares_error_minimum_percent_change(self) -> float:
        return self.__sum_of_squares_error_minimum_percent_change
    
    @property
    def random_seed(self) -> int:
        return self.__random_seed
    
    @bldg_ids.setter
    def bldg_ids(self, value: List[int]):
        assert len(value) > self.__MINIMUM_BUILDING_COUNT, f'Provide at least {self.__MINIMUM_BUILDING_COUNT} bldg_ids.'
        self.__bldg_ids = sorted(value)
    
    @maximum_clusters.setter
    def maximum_clusters(self, value: int):
        value = math.ceil(len(self.bldg_ids)/2) if value is None else value
        assert 2 <= value < len(self.bldg_ids), f'maximum_clusters must be > 2 and less than number of bldg_ids'
        self.__maximum_clusters = value

    @sum_of_squares_error_minimum_percent_change.setter
    def sum_of_squares_error_minimum_percent_change(self, value: float):
        self.__sum_of_squares_error_minimum_percent_change = 10.0 if value is None else value

    @random_seed.setter
    def random_seed(self, value: int):
        self.__random_seed = random.randint(*Environment.DEFAULT_RANDOM_SEED_RANGE) if value is None else value

    def cluster(self) -> Tuple[int, pd.DataFrame, pd.DataFrame]:
        data = self.set_data()
        scaler = MinMaxScaler()
        scaler.fit(data)
        data = scaler.transform(data)
        scores = []
        labels = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = {executor.submit(self.__cluster, *(data, c)): c for c in range(2, self.maximum_clusters + 1)}
        
            for future in concurrent.futures.as_completed(results):
                try:
                    clusters = results[future]
                    _labels, sum_of_squares_error, calinski_harabasz_score, silhouette_score, davies_bouldin_score = future.result()
                    scores.append({
                        'clusters': clusters,
                        'sum_of_square_error': sum_of_squares_error,
                        'calinski_harabasz_score': calinski_harabasz_score,
                        'silhouette_score': silhouette_score,
                        'davies_bouldin_score': davies_bouldin_score
                    })
                    labels.append(pd.DataFrame({
                        'clusters': clusters,
                        'bldg_id': self.bldg_ids,
                        'label': _labels
                    }))

                except Exception as e:
                    raise(e)

        scores = pd.DataFrame(scores).sort_values('clusters')
        labels = pd.concat(labels, ignore_index=True)
        optimal_clusters = self.get_optimal_clusters(*scores.T.values)

        return optimal_clusters, scores, labels
    
    def get_optimal_clusters(self, clusters: List[int], sum_of_squares_error: List[float], calinski_harabasz_score: List[float], silhouette_score: List[float], davies_bouldin_score: List[float]) -> int:
        assert len(clusters) == len(sum_of_squares_error) == len(calinski_harabasz_score) == len(davies_bouldin_score), \
            'clusters and scores lists must have equal lengths.'
        sum_of_squares_error_change = (
            np.array(sum_of_squares_error, dtype=float)[:-1] 
            - np.array(sum_of_squares_error, dtype=float)[1:]
        )*100/np.array(sum_of_squares_error, dtype=float)[:-1]
        sse_candidates = np.array(clusters[:-1], dtype=int)[
            sum_of_squares_error_change < self.sum_of_squares_error_minimum_percent_change
        ]
        optimal_clusters = np.nanmean([
            # sse_candidates.min() if len(sse_candidates) > 0 else np.nan,
            clusters[np.array(calinski_harabasz_score, dtype=float).argmax()],
            clusters[np.array(silhouette_score, dtype=float).argmax()],
            clusters[np.array(davies_bouldin_score, dtype=float).argmin()],

        ], dtype=float)
        optimal_clusters = math.floor(optimal_clusters)
        
        return optimal_clusters

    def __cluster(self, data: np.ndarray, clusters: int):
        model = KMeans(clusters, random_state=self.random_seed).fit(data)
        labels = model.labels_
        sum_of_squares_error = model.inertia_
        calinski_harabasz_score = cluster_metrics.calinski_harabasz_score(data, model.labels_)
        davies_bouldin_score = cluster_metrics.davies_bouldin_score(data, model.labels_)
        silhouette_score = cluster_metrics.silhouette_score(data, model.labels_)

        return labels, sum_of_squares_error, calinski_harabasz_score, silhouette_score, davies_bouldin_score
    
    def set_data(self) -> pd.DataFrame:
        raise NotImplementedError

class MetadataClustering(Clustering):
    def __init__(self, end_use_load_profiles: EndUseLoadProfiles, bldg_ids: List[int], maximum_clusters: int = None, sum_of_squares_error_minimum_percent_change: float = None, random_seed: int = None):
        super().__init__(
            end_use_load_profiles, 
            bldg_ids, 
            maximum_clusters=maximum_clusters,
            sum_of_squares_error_minimum_percent_change=sum_of_squares_error_minimum_percent_change,
            random_seed=random_seed,
        )

    def set_data(self) -> pd.DataFrame:
        data = self.end_use_load_profiles.metadata.metadata.get()
        data = data[data.index.isin(self.bldg_ids)].copy()
        clustering_columns = [
            'in.window_to_wall_ratio',
            'in.vintage',
            'in.orientation_hour_sin',
            'in.orientation_hour_cos',
            'in.infiltration',
            'in.insulation_ceiling',
            'in.insulation_slab',
            'in.insulation_wall',
            'out.site_energy.total.energy_consumption_intensity',
        ]

        # window-to-wall ratio
        data['in.window_to_wall_ratio'] = data['in.window_area_ft_2']/data['in.wall_area_above_grade_exterior_ft_2']

        # vintage
        data['in.vintage'] = data['in.vintage'].replace('\D', '', regex=True).astype(int)

        # orientation: cosine transformation
        order = ['North', 'Northeast', 'East', 'Southeast', 'South', 'Southwest', 'West', 'Northwest']
        data['in.orientation'] = data['in.orientation'].map(lambda x: order.index(x))
        data['in.orientation_hour_sin'] = np.sin(2 * np.pi * data['in.orientation']/(len(order)-1))
        data['in.orientation_hour_cos'] = np.cos(2 * np.pi * data['in.orientation']/(len(order)-1))
        data = data.drop(columns=['in.orientation'])

        # infiltration
        data['in.infiltration'] = data['in.infiltration'].replace(regex=r' ACH50', value='')
        data['in.infiltration'] = pd.to_numeric(data['in.infiltration'], errors='coerce')
        data['in.infiltration'] = data['in.infiltration'].fillna(0)

        # insulation ceiling
        data['in.insulation_ceiling'] = data['in.insulation_ceiling'].replace(regex=r'R-', value='')
        data['in.insulation_ceiling'] = pd.to_numeric(data['in.insulation_ceiling'], errors='coerce')
        data['in.insulation_ceiling'] = data['in.insulation_ceiling'].fillna(0)

        # insulation slab
        data.loc[data['in.insulation_slab'].isin(['None', 'Uninsulated']), 'in.insulation_slab'] = 'Uninsulated R0'
        data['in.insulation_slab'] = data['in.insulation_slab'].str.split(' ', expand=True)[1]
        data['in.insulation_slab'] = data['in.insulation_slab'].replace(regex=r'R', value='')
        data['in.insulation_slab'] = pd.to_numeric(data['in.insulation_slab'], errors='coerce')
        data['in.insulation_slab'] = data['in.insulation_slab'].fillna(0)

        # insulation wall
        data['in.insulation_wall'] = data['in.insulation_wall'].replace(regex=r'.+ R-', value='')
        data['in.insulation_wall'] = pd.to_numeric(data['in.insulation_wall'], errors='coerce')
        data['in.insulation_wall'] = data['in.insulation_wall'].fillna(0)

        data = data[clustering_columns].astype(float)

        return data